# Copyright (c) 2022 elParaguayo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import shutil

import pytest

import libqtile.bar
import libqtile.config
from libqtile import widget
from test.helpers import Retry


def test_no_duplicates_multiple_instances(manager_nospawn, minimal_conf_noscreen, backend_name):
    """Check only one instance of Systray widget."""
    if backend_name == "wayland":
        pytest.skip("Skipping test on Wayland.")

    assert not widget.Systray._instances
    config = minimal_conf_noscreen
    config.screens = [
        libqtile.config.Screen(top=libqtile.bar.Bar([widget.Systray(), widget.Systray()], 10))
    ]

    manager_nospawn.start(config)

    widgets = manager_nospawn.c.bar["top"].info()["widgets"]
    assert len(widgets) == 2
    assert widgets[1]["name"] == "configerrorwidget"


def test_no_duplicates_mirror(manager_nospawn, minimal_conf_noscreen, backend_name):
    """Check systray is not mirrored."""
    if backend_name == "wayland":
        pytest.skip("Skipping test on Wayland.")

    assert not widget.Systray._instances
    systray = widget.Systray()
    config = minimal_conf_noscreen
    config.fake_screens = [
        libqtile.config.Screen(
            top=libqtile.bar.Bar([systray], 10),
            x=0,
            y=0,
            width=300,
            height=300,
        ),
        libqtile.config.Screen(
            top=libqtile.bar.Bar([systray], 10),
            x=0,
            y=300,
            width=300,
            height=300,
        ),
    ]

    manager_nospawn.start(config)

    # Second screen has tried to mirror the Systray instance
    widgets = manager_nospawn.c.screen[1].bar["top"].info()["widgets"]
    assert len(widgets) == 1
    assert widgets[0]["name"] == "configerrorwidget"


def test_systray_reconfigure_screens(manager_nospawn, minimal_conf_noscreen, backend_name):
    """Check systray does not crash when reconfiguring screens."""
    if backend_name == "wayland":
        pytest.skip("Skipping test on Wayland.")

    assert not widget.Systray._instances
    config = minimal_conf_noscreen
    config.screens = [libqtile.config.Screen(top=libqtile.bar.Bar([widget.Systray()], 10))]

    manager_nospawn.start(config)

    assert manager_nospawn.c.bar["top"].info()["widgets"][0]["name"] == "systray"

    manager_nospawn.c.reconfigure_screens()

    assert manager_nospawn.c.bar["top"].info()["widgets"][0]["name"] == "systray"


def test_systray_icons(manager_nospawn, minimal_conf_noscreen, backend_name):
    """Check icons are placed correctly."""

    @Retry(ignore_exceptions=(AssertionError))
    def wait_for_icons():
        _, icons = manager_nospawn.c.widget["systray"].eval("len(self.tray_icons)")
        assert int(icons) == 2

    if backend_name == "wayland":
        pytest.skip("Skipping test on Wayland.")

    for prog in ("volumeicon", "vlc"):
        if shutil.which(prog) is None:
            pytest.skip(f"{prog} must be installed. Skipping test.")

    config = minimal_conf_noscreen
    config.screens = [libqtile.config.Screen(top=libqtile.bar.Bar([widget.Systray()], 40))]

    manager_nospawn.start(config)

    # No icons at this stage so length is 0
    assert manager_nospawn.c.widget["systray"].info()["widget"]["length"] == 0

    manager_nospawn.c.spawn("volumeicon")
    manager_nospawn.c.spawn("vlc")

    wait_for_icons()

    # We now have two icons so widget should expand to leave space in bar for icons
    assert manager_nospawn.c.widget["systray"].info()["widget"]["length"] > 0

    # Check positioning of icon
    _, x = manager_nospawn.c.widget["systray"].eval("self.tray_icons[0].x")
    _, y = manager_nospawn.c.widget["systray"].eval("self.tray_icons[0].y")

    # Positions are relative to bar
    assert (int(x), int(y)) == (3, 10)

    # Icons should be in alphabetical order
    _, order = manager_nospawn.c.widget["systray"].eval("[i.name for i in self.tray_icons]")

    assert order == "['vlc', 'volumeicon']"



# Copyright (c) 2021 Matt Colligan
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import abc
import functools
import typing

import cairocffi
import pywayland
import wlroots.wlr_types.foreign_toplevel_management_v1 as ftm
from pywayland.server import Listener
from wlroots import PtrHasData, ffi, xwayland
from wlroots.util.box import Box
from wlroots.util.edges import Edges
from wlroots.wlr_types import Texture
from wlroots.wlr_types.idle_inhibit_v1 import IdleInhibitorV1
from wlroots.wlr_types.layer_shell_v1 import LayerShellV1Layer, LayerSurfaceV1
from wlroots.wlr_types.pointer_constraints_v1 import (
    PointerConstraintV1,
    PointerConstraintV1StateField,
)
from wlroots.wlr_types.xdg_shell import XdgPopup, XdgSurface, XdgTopLevelSetFullscreenEvent

from libqtile import config, hook, utils
from libqtile.backend import base
from libqtile.backend.base import FloatStates
from libqtile.backend.wayland.drawer import Drawer
from libqtile.backend.wayland.wlrq import DRM_FORMAT_ARGB8888, HasListeners
from libqtile.command.base import CommandError
from libqtile.log_utils import logger

if typing.TYPE_CHECKING:
    from typing import Any

    from wlroots.wlr_types.surface import SubSurface as WlrSubSurface
    from wlroots.wlr_types.surface import Surface

    from libqtile.backend.wayland.core import Core
    from libqtile.backend.wayland.output import Output
    from libqtile.command.base import CommandObject, ItemT
    from libqtile.core.manager import Qtile
    from libqtile.group import _Group
    from libqtile.utils import ColorsType, ColorType

EDGES_TILED = Edges.TOP | Edges.BOTTOM | Edges.LEFT | Edges.RIGHT
EDGES_FLOAT = Edges.NONE

S = typing.TypeVar("S", bound=PtrHasData)


@functools.lru_cache()
def _rgb(color: ColorType) -> ffi.CData:
    """Helper to create and cache float[4] arrays for border painting"""
    if isinstance(color, ffi.CData):
        return color
    return ffi.new("float[4]", utils.rgb(color))


class _Base:
    _wid: int

    @property
    def wid(self) -> int:
        return self._wid

    @property
    def width(self) -> int:
        return self._width

    @width.setter
    def width(self, width: int) -> None:
        self._width = width

    @property
    def height(self) -> int:
        return self._height

    @height.setter
    def height(self, height: int) -> None:
        self._height = height


class Window(typing.Generic[S], _Base, base.Window, HasListeners):
    """
    This is a generic window class for "regular" windows. The type variable `S` denotes
    which type of surface the window manages, and by extension which shell the window
    belongs to. While this does implement some of `base.Window`'s abstract methods, the
    concrete classes are responsible for implementing a few others.
    """

    def __init__(self, core: Core, qtile: Qtile, surface: S):
        base.Window.__init__(self)
        self.core = core
        self.qtile = qtile
        self.surface = surface
        self._group: _Group | None = None
        self._mapped: bool = False
        self.x = 0
        self.y = 0
        self.bordercolor: list[ffi.CData] = [_rgb((0, 0, 0, 1))]
        self._opacity: float = 1.0
        self._outputs: set[Output] = set()
        self._wm_class: str | None = None
        self._idle_inhibitors_count: int = 0

        self._width: int = 0
        self._height: int = 0
        self.float_x: int | None = None
        self.float_y: int | None = None
        self._float_width: int = 0
        self._float_height: int = 0
        self._float_state = FloatStates.NOT_FLOATING

        surface.data = self.ftm_handle = core.foreign_toplevel_manager_v1.create_handle()

    def finalize(self) -> None:
        self.finalize_listeners()
        self.ftm_handle.destroy()

    @property
    def wid(self) -> int:
        return self._wid

    @property
    def group(self) -> _Group | None:
        return self._group

    @group.setter
    def group(self, group: _Group | None) -> None:
        self._group = group

    @property
    def mapped(self) -> bool:
        return self._mapped

    @mapped.setter
    def mapped(self, mapped: bool) -> None:
        """We keep track of which windows are mapped so we know which to render"""
        if mapped == self._mapped:
            return
        self._mapped = mapped
        if mapped:
            self.core.mapped_windows.append(self)
        else:
            self.core.mapped_windows.remove(self)
        self.core.stack_windows()
        if self._idle_inhibitors_count > 0:
            self.core.check_idle_inhibitor()

    def _on_destroy(self, _listener: Listener, _data: Any) -> None:
        logger.debug("Signal: window destroy")
        if self.mapped:
            logger.warning("Window destroyed before unmap event.")
            self.mapped = False

        if self in self.core.pending_windows:
            self.core.pending_windows.remove(self)
        else:
            self.qtile.unmanage(self.wid)

        self.finalize()

    def _on_commit(self, _listener: Listener, _data: Any) -> None:
        self.damage()

    def _on_foreign_request_maximize(
        self, _listener: Listener, event: ftm.ForeignToplevelHandleV1MaximizedEvent
    ) -> None:
        logger.debug("Signal: foreign_toplevel_management request_maximize")
        self.maximized = event.maximized

    def _on_foreign_request_minimize(
        self, _listener: Listener, event: ftm.ForeignToplevelHandleV1MinimizedEvent
    ) -> None:
        logger.debug("Signal: foreign_toplevel_management request_minimize")
        self.minimized = event.minimized

    def _on_foreign_request_fullscreen(
        self, _listener: Listener, event: ftm.ForeignToplevelHandleV1FullscreenEvent
    ) -> None:
        logger.debug("Signal: foreign_toplevel_management request_fullscreen")
        self.fullscreen = event.fullscreen

    def _on_foreign_request_activate(
        self, _listener: Listener, event: ftm.ForeignToplevelHandleV1ActivatedEvent
    ) -> None:
        logger.debug("Signal: foreign_toplevel_management request_activate")
        if self.group:
            self.qtile.current_screen.set_group(self.group)
            self.group.focus(self)

    def _on_foreign_request_close(self, _listener: Listener, _data: Any) -> None:
        logger.debug("Signal: foreign_toplevel_management request_close")
        self.kill()

    def _on_inhibitor_destroy(self, listener: Listener, surface: Surface) -> None:
        # We don't have reference to the inhibitor, but it doesn't really
        # matter we only need to keep count of how many inhibitors there are
        self._idle_inhibitors_count -= 1
        listener.remove()
        if self._idle_inhibitors_count == 0:
            self.core.check_idle_inhibitor()

    def _find_outputs(self) -> None:
        """Find the outputs on which this window can be seen."""
        self._outputs = set(o for o in self.core.outputs if o.contains(self))

    def damage(self) -> None:
        for output in self._outputs:
            output.damage()

    def get_wm_class(self) -> list | None:
        if self._wm_class:
            return [self._wm_class]
        return None

    def focus(self, warp: bool) -> None:
        self.core.focus_window(self)

        if warp and self.qtile.config.cursor_warp:
            self.core.warp_pointer(
                self.x + self.width / 2,
                self.y + self.height / 2,
            )

        if self.group:
            self.group.current_window = self

        hook.fire("client_focus", self)

    def togroup(
        self, group_name: str | None = None, *, switch_group: bool = False, toggle: bool = False
    ) -> None:
        """
        Move window to a specified group

        Also switch to that group if switch_group is True.

        If `toggle` is True and and the specified group is already on the screen,
        use the last used group as target instead.
        """
        if group_name is None:
            group = self.qtile.current_group
        else:
            if group_name not in self.qtile.groups_map:
                raise CommandError("No such group: %s" % group_name)
            group = self.qtile.groups_map[group_name]

        if self.group is group:
            if toggle and hasattr(self.group.screen, "previous_group"):
                group = self.group.screen.previous_group
            else:
                return

        self.hide()
        if self.group:
            if self.group.screen:
                # for floats remove window offset
                self.x -= self.group.screen.x
            self.group.remove(self)

        if group.screen and self.x < group.screen.x:
            self.x += group.screen.x
        group.add(self)
        if switch_group:
            group.cmd_toscreen(toggle=toggle)

    def paint_borders(self, color: ColorsType | None, width: int) -> None:
        if color:
            if isinstance(color, list):
                if len(color) > width:
                    color = color[:width]
                self.bordercolor = [_rgb(c) for c in color]
            else:
                self.bordercolor = [_rgb(color)]
        self.borderwidth = width

    @property
    def floating(self) -> bool:
        return self._float_state != FloatStates.NOT_FLOATING

    @floating.setter
    def floating(self, do_float: bool) -> None:
        if do_float and self._float_state == FloatStates.NOT_FLOATING:
            if self.group and self.group.screen:
                screen = self.group.screen
                if not self._float_width:  # These might start as 0
                    self._float_width = self.width
                    self._float_height = self.height
                self._reconfigure_floating(
                    screen.x + self.float_x,
                    screen.y + self.float_y,
                    self._float_width,
                    self._float_height,
                )
            else:
                # if we are setting floating early, e.g. from a hook, we don't have a screen yet
                self._float_state = FloatStates.FLOATING
        elif (not do_float) and self._float_state != FloatStates.NOT_FLOATING:
            self._update_fullscreen(False)
            if self._float_state == FloatStates.FLOATING:
                # store last size
                self._float_width = self.width
                self._float_height = self.height
            self._float_state = FloatStates.NOT_FLOATING
            if self.group:
                self.group.mark_floating(self, False)
            hook.fire("float_change")

    @abc.abstractmethod
    def _update_fullscreen(self, do_full: bool) -> None:
        pass

    @property
    def maximized(self) -> bool:
        return self._float_state == FloatStates.MAXIMIZED

    @maximized.setter
    def maximized(self, do_maximize: bool) -> None:
        if do_maximize:
            screen = (self.group and self.group.screen) or self.qtile.find_closest_screen(
                self.x, self.y
            )
            bw = self.group.floating_layout.max_border_width if self.group else 0
            self._reconfigure_floating(
                screen.dx,
                screen.dy,
                screen.dwidth - 2 * bw,
                screen.dheight - 2 * bw,
                new_float_state=FloatStates.MAXIMIZED,
            )
        else:
            if self._float_state == FloatStates.MAXIMIZED:
                self.floating = False

        self.ftm_handle.set_maximized(do_maximize)

    @property
    def minimized(self) -> bool:
        return self._float_state == FloatStates.MINIMIZED

    @minimized.setter
    def minimized(self, do_minimize: bool) -> None:
        if do_minimize:
            if self._float_state != FloatStates.MINIMIZED:
                self._reconfigure_floating(new_float_state=FloatStates.MINIMIZED)
        else:
            if self._float_state == FloatStates.MINIMIZED:
                self.floating = False

        self.ftm_handle.set_minimized(do_minimize)

    def _tweak_float(
        self,
        x: int | None = None,
        y: int | None = None,
        dx: int = 0,
        dy: int = 0,
        w: int | None = None,
        h: int | None = None,
        dw: int = 0,
        dh: int = 0,
    ) -> None:
        if x is None:
            x = self.x
        x += dx

        if y is None:
            y = self.y
        y += dy

        if w is None:
            w = self.width
        w += dw

        if h is None:
            h = self.height
        h += dh

        if h < 0:
            h = 0
        if w < 0:
            w = 0

        screen = self.qtile.find_closest_screen(x + w // 2, y + h // 2)
        if self.group and screen is not None and screen != self.group.screen:
            self.group.remove(self, force=True)
            screen.group.add(self, force=True)
            self.qtile.focus_screen(screen.index)

        self._reconfigure_floating(x, y, w, h)

    def _reconfigure_floating(
        self,
        x: int | None = None,
        y: int | None = None,
        w: int | None = None,
        h: int | None = None,
        new_float_state: FloatStates = FloatStates.FLOATING,
    ) -> None:
        self._update_fullscreen(new_float_state == FloatStates.FULLSCREEN)
        if new_float_state == FloatStates.MINIMIZED:
            self.hide()
        else:
            self.place(
                x, y, w, h, self.borderwidth, self.bordercolor, above=True, respect_hints=True
            )
        if self._float_state != new_float_state:
            self._float_state = new_float_state
            if self.group:  # may be not, if it's called from hook
                self.group.mark_floating(self, True)
            hook.fire("float_change")

    def info(self) -> dict:
        """Return a dictionary of info."""
        float_info = {
            "x": self.float_x,
            "y": self.float_y,
            "width": self._float_width,
            "height": self._float_height,
        }
        return dict(
            name=self.name,
            x=self.x,
            y=self.y,
            width=self.width,
            height=self.height,
            group=self.group.name if self.group else None,
            id=self.wid,
            wm_class=self.get_wm_class(),
            float_info=float_info,
            floating=self._float_state != FloatStates.NOT_FLOATING,
            maximized=self._float_state == FloatStates.MAXIMIZED,
            minimized=self._float_state == FloatStates.MINIMIZED,
            fullscreen=self._float_state == FloatStates.FULLSCREEN,
        )

    def match(self, match: config.Match) -> bool:
        return match.compare(self)

    def add_idle_inhibitor(
        self, surface: Surface, _x: int, _y: int, inhibitor: IdleInhibitorV1 | None
    ) -> None:
        if inhibitor and surface == inhibitor.surface:
            self._idle_inhibitors_count += 1
            inhibitor.data = self
            self.add_listener(inhibitor.destroy_event, self._on_inhibitor_destroy)
            if self._idle_inhibitors_count == 1:
                self.core.check_idle_inhibitor()

    @property
    def is_idle_inhibited(self) -> bool:
        return self._idle_inhibitors_count > 0

    def _items(self, name: str) -> ItemT:
        if name == "group":
            return True, []
        if name == "layout":
            if self.group:
                return True, list(range(len(self.group.layouts)))
            return None
        if name == "screen":
            if self.group and self.group.screen:
                return True, []
        return None

    def _select(self, name: str, sel: str | int | None) -> CommandObject | None:
        if name == "group":
            return self.group
        elif name == "layout":
            if sel is None:
                return self.group.layout if self.group else None
            else:
                return utils.lget(self.group.layouts, sel) if self.group else None
        elif name == "screen":
            return self.group.screen if self.group else None
        return None

    def cmd_focus(self, warp: bool = True) -> None:
        """Focuses the window."""
        self.focus(warp)

    def cmd_move_floating(self, dx: int, dy: int) -> None:
        self._tweak_float(dx=dx, dy=dy)

    def cmd_resize_floating(self, dw: int, dh: int) -> None:
        self._tweak_float(dw=dw, dh=dh)

    def cmd_set_position_floating(self, x: int, y: int) -> None:
        self._tweak_float(x=x, y=y)

    def cmd_set_position(self, x: int, y: int) -> None:
        if self.floating:
            self._tweak_float(x=x, y=y)
            return

        if self.group:
            cx = self.core.cursor.x
            cy = self.core.cursor.y
            for window in self.group.windows:
                if (
                    window is not self
                    and not window.floating
                    and window.x <= cx <= (window.x + window.width)
                    and window.y <= cy <= (window.y + window.height)
                ):
                    clients = self.group.layout.clients
                    index1 = clients.index(self)
                    index2 = clients.index(window)
                    clients[index1], clients[index2] = clients[index2], clients[index1]
                    self.group.layout.focused = index2
                    self.group.layout_all()
                    return

    def cmd_set_size_floating(self, w: int, h: int) -> None:
        self._tweak_float(w=w, h=h)

    def cmd_place(
        self,
        x: int,
        y: int,
        width: int,
        height: int,
        borderwidth: int,
        bordercolor: ColorsType | None,
        above: bool = False,
        margin: int | None = None,
    ) -> None:
        self.place(x, y, width, height, borderwidth, bordercolor, above, margin)

    def cmd_get_position(self) -> tuple[int, int]:
        return self.x, self.y

    def cmd_get_size(self) -> tuple[int, int]:
        return self.width, self.height

    def cmd_toggle_floating(self) -> None:
        self.floating = not self.floating

    def cmd_enable_floating(self) -> None:
        self.floating = True

    def cmd_disable_floating(self) -> None:
        self.floating = False

    def cmd_toggle_maximize(self) -> None:
        self.maximized = not self.maximized

    def cmd_toggle_minimize(self) -> None:
        self.minimized = not self.minimized

    def cmd_toggle_fullscreen(self) -> None:
        self.fullscreen = not self.fullscreen

    def cmd_enable_fullscreen(self) -> None:
        self.fullscreen = True

    def cmd_disable_fullscreen(self) -> None:
        self.fullscreen = False

    def cmd_bring_to_front(self) -> None:
        if self.mapped:
            self.core.mapped_windows.remove(self)
            self.core.mapped_windows.append(self)
            self.core.stack_windows()

    def cmd_kill(self) -> None:
        self.kill()

    def cmd_static(
        self,
        screen: int | None = None,
        x: int | None = None,
        y: int | None = None,
        width: int | None = None,
        height: int | None = None,
    ) -> None:
        # The concrete Window type must fire the client_managed hook after it's
        # completed any custom logic.
        self.defunct = True
        if self.group:
            self.group.remove(self)
        if x is None:
            x = self.x + self.borderwidth
        if y is None:
            y = self.y + self.borderwidth
        if width is None:
            width = self.width
        if height is None:
            height = self.height

        self.finalize_listeners()
        win = self._to_static()
        if screen is not None:
            win.screen = self.qtile.screens[screen]
        win.mapped = True
        win.place(x, y, width, height, 0, None)
        self.qtile.windows_map[self.wid] = win
        if self in self.core.mapped_windows:
            self.core.mapped_windows.remove(self)
        self.core.stack_windows()

    @abc.abstractmethod
    def _to_static(self) -> Static:
        # This must return a new `base.Static` subclass instance
        pass


class XdgWindow(Window[XdgSurface]):
    """An Wayland client connecting via the xdg shell."""

    def __init__(self, core: Core, qtile: Qtile, surface: XdgSurface):
        Window.__init__(self, core, qtile, surface)

        self._wm_class = surface.toplevel.app_id
        self.popups: list[XdgPopupWindow] = []
        self.subsurfaces: list[SubSurface] = []

        self.add_listener(surface.map_event, self._on_map)
        self.add_listener(surface.unmap_event, self._on_unmap)
        self.add_listener(surface.destroy_event, self._on_destroy)
        self.add_listener(surface.new_popup_event, self._on_new_popup)
        self.add_listener(surface.surface.commit_event, self._on_commit)
        self.add_listener(surface.surface.new_subsurface_event, self._on_new_subsurface)

    def finalize(self) -> None:
        Window.finalize(self)
        for subsurface in self.subsurfaces:
            subsurface.finalize()

        for pc in self.core.pointer_constraints.copy():
            if pc.window is self:
                pc.finalize()

    def _on_map(self, _listener: Listener, _data: Any) -> None:
        logger.debug("Signal: xdgwindow map")

        if self in self.core.pending_windows:
            self.core.pending_windows.remove(self)
            self._wid = self.core.new_wid()
            logger.debug(f"Managing new top-level window with window ID: {self.wid}")

            # Save the client's desired geometry
            surface = self.surface
            geometry = surface.get_geometry()
            self._width = self._float_width = geometry.width
            self._height = self._float_height = geometry.height

            # Tell the client to render tiled edges
            surface.set_tiled(EDGES_TILED)

            # Get the client's name
            if surface.toplevel.title:
                self.name = surface.toplevel.title
                self.ftm_handle.set_title(self.name)
            if self._wm_class:
                self.ftm_handle.set_app_id(self._wm_class or "")

            # Add the toplevel's listeners
            self.add_listener(
                surface.toplevel.request_fullscreen_event, self._on_request_fullscreen
            )
            self.add_listener(surface.toplevel.set_title_event, self._on_set_title)
            self.add_listener(surface.toplevel.set_app_id_event, self._on_set_app_id)
            self.add_listener(
                self.ftm_handle.request_maximize_event, self._on_foreign_request_maximize
            )
            self.add_listener(
                self.ftm_handle.request_minimize_event, self._on_foreign_request_minimize
            )
            self.add_listener(
                self.ftm_handle.request_activate_event, self._on_foreign_request_activate
            )
            self.add_listener(
                self.ftm_handle.request_fullscreen_event, self._on_foreign_request_fullscreen
            )
            self.add_listener(self.ftm_handle.request_close_event, self._on_foreign_request_close)

            self.qtile.manage(self)

        if self.group and self.group.screen:
            self.mapped = True
            self.core.focus_window(self)

    def _on_unmap(self, _listener: Listener, _data: Any) -> None:
        logger.debug("Signal: xdgwindow unmap")
        self.mapped = False
        self.damage()
        seat = self.core.seat
        if not seat.destroyed:
            if self.surface.surface == seat.keyboard_state.focused_surface:
                seat.keyboard_clear_focus()

    def _on_new_popup(self, _listener: Listener, xdg_popup: XdgPopup) -> None:
        logger.debug("Signal: xdgwindow new_popup")
        self.popups.append(XdgPopupWindow(self, xdg_popup))

    def _on_new_subsurface(self, _listener: Listener, subsurface: WlrSubSurface) -> None:
        self.subsurfaces.append(SubSurface(self, subsurface))

    def _on_request_fullscreen(
        self, _listener: Listener, event: XdgTopLevelSetFullscreenEvent
    ) -> None:
        logger.debug("Signal: xdgwindow request_fullscreen")
        if self.qtile.config.auto_fullscreen:
            self.fullscreen = event.fullscreen

    def _on_set_title(self, _listener: Listener, _data: Any) -> None:
        logger.debug("Signal: xdgwindow set_title")
        title = self.surface.toplevel.title
        if title and title != self.name:
            self.name = title
            self.ftm_handle.set_title(self.name)
            hook.fire("client_name_updated", self)

    def _on_set_app_id(self, _listener: Listener, _data: Any) -> None:
        logger.debug("Signal: xdgwindow set_app_id")
        self._wm_class = self.surface.toplevel.app_id
        self.ftm_handle.set_app_id(self._wm_class or "")

    def hide(self) -> None:
        if self.mapped:
            self.surface.unmap_event.emit()

    def unhide(self) -> None:
        if not self.mapped:
            self.surface.map_event.emit()

    def kill(self) -> None:
        self.surface.send_close()

    def has_fixed_size(self) -> bool:
        state = self.surface.toplevel._ptr.current
        return 0 < state.min_width == state.max_width and 0 < state.min_height == state.max_height

    def is_transient_for(self) -> base.WindowType | None:
        """What window is this window a transient window for?"""
        parent = self.surface.toplevel.parent
        if parent:
            for win in self.qtile.windows_map.values():
                if isinstance(win, XdgWindow) and win.surface == parent:
                    return win
        return None

    def get_pid(self) -> int:
        pid = pywayland.ffi.new("pid_t *")
        pywayland.lib.wl_client_get_credentials(
            self.surface._ptr.client.client, pid, ffi.NULL, ffi.NULL
        )
        return pid[0]

    def _update_fullscreen(self, do_full: bool) -> None:
        if do_full != (self._float_state == FloatStates.FULLSCREEN):
            self.surface.set_fullscreen(do_full)
            self.ftm_handle.set_fullscreen(do_full)

    @property
    def fullscreen(self) -> bool:
        return self._float_state == FloatStates.FULLSCREEN

    @fullscreen.setter
    def fullscreen(self, do_full: bool) -> None:
        self.surface.set_fullscreen(do_full)
        if do_full:
            screen = (self.group and self.group.screen) or self.qtile.find_closest_screen(
                self.x, self.y
            )
            if self.group:
                bw = self.group.floating_layout.fullscreen_border_width
            else:
                bw = 0
            self._reconfigure_floating(
                screen.x,
                screen.y,
                screen.width - 2 * bw,
                screen.height - 2 * bw,
                new_float_state=FloatStates.FULLSCREEN,
            )
        elif self._float_state == FloatStates.FULLSCREEN:
            self.floating = False

        self.ftm_handle.set_fullscreen(do_full)

    def place(
        self,
        x: int,
        y: int,
        width: int,
        height: int,
        borderwidth: int,
        bordercolor: ColorsType | None,
        above: bool = False,
        margin: int | list[int] | None = None,
        respect_hints: bool = False,
    ) -> None:
        # Adjust the placement to account for layout margins, if there are any.
        if margin is not None:
            if isinstance(margin, int):
                margin = [margin] * 4
            x += margin[3]
            y += margin[0]
            width -= margin[1] + margin[3]
            height -= margin[0] + margin[2]

        if respect_hints:
            state = self.surface.toplevel._ptr.current
            width = max(width, state.min_width)
            height = max(height, state.min_height)
            if state.max_width:
                width = min(width, state.max_width)
            if state.max_height:
                height = min(height, state.max_height)

        # save x and y float offset
        if self.group is not None and self.group.screen is not None:
            self.float_x = x - self.group.screen.x
            self.float_y = y - self.group.screen.y

        self.x = x
        self.y = y
        self._width = int(width)
        self._height = int(height)
        self.surface.set_size(self._width, self._height)
        self.paint_borders(bordercolor, borderwidth)

        if above:
            self.cmd_bring_to_front()

        prev_outputs = self._outputs.copy()
        self._find_outputs()
        for output in self._outputs | prev_outputs:
            output.damage()

    def cmd_static(
        self,
        screen: int | None = None,
        x: int | None = None,
        y: int | None = None,
        width: int | None = None,
        height: int | None = None,
    ) -> None:
        Window.cmd_static(self, screen, x, y, width, height)
        win = self.qtile.windows_map[self._wid]
        assert isinstance(win, XdgStatic)
        win.subsurfaces = self.subsurfaces

        for pc in self.core.pointer_constraints.copy():
            if pc.window is self:
                pc.window = win
                break

        hook.fire("client_managed", win)

    def _to_static(self) -> XdgStatic:
        return XdgStatic(
            self.core, self.qtile, self.surface, self.wid, self._idle_inhibitors_count
        )


class XWindow(Window[xwayland.Surface]):
    """An X11 client connecting via XWayland."""

    def __init__(self, core: Core, qtile: Qtile, surface: xwayland.Surface):
        Window.__init__(self, core, qtile, surface)

        self._wm_class = self.surface.wm_class
        self._unmapping: bool = False  # Whether the client or Qtile unmapped this

        self.add_listener(surface.map_event, self._on_map)
        self.add_listener(surface.unmap_event, self._on_unmap)
        self.add_listener(surface.destroy_event, self._on_destroy)

    def _on_map(self, _listener: Listener, _data: Any) -> None:
        logger.debug("Signal: xwindow map")

        if self in self.core.pending_windows:
            self.core.pending_windows.remove(self)
            self._wid = self.core.new_wid()
            logger.debug(f"Managing new XWayland window with window ID: {self.wid}")

            # Make it static if it isn't a regular window
            if self.surface.override_redirect:
                self.cmd_static(
                    None, self.surface.x, self.surface.y, self.surface.width, self.surface.height
                )
                win = self.qtile.windows_map[self._wid]
                assert isinstance(win, XStatic)
                self.core.focus_window(win)
                return

            # Save the client's desired geometry. xterm seems to have these set to 1, so
            # let's ignore 1 or below. The float sizes will be fetched when it is floated.
            if self.surface.width > 1:
                self._width = self._float_width = self.surface.width
            if self.surface.height > 1:
                self._height = self._float_height = self.surface.height

            # Get the client's name and class
            title = self.surface.title
            if title:
                self.name = title
                self.ftm_handle.set_title(self.name)
            self._wm_class = self.surface.wm_class
            self.ftm_handle.set_app_id(self._wm_class or "")

            # Add event listeners
            self.add_listener(self.surface.surface.commit_event, self._on_commit)
            self.add_listener(self.surface.request_fullscreen_event, self._on_request_fullscreen)
            self.add_listener(self.surface.set_title_event, self._on_set_title)
            self.add_listener(self.surface.set_class_event, self._on_set_class)
            self.add_listener(
                self.ftm_handle.request_maximize_event, self._on_foreign_request_maximize
            )
            self.add_listener(
                self.ftm_handle.request_minimize_event, self._on_foreign_request_minimize
            )
            self.add_listener(
                self.ftm_handle.request_activate_event, self._on_foreign_request_activate
            )
            self.add_listener(
                self.ftm_handle.request_fullscreen_event, self._on_foreign_request_fullscreen
            )
            self.add_listener(self.ftm_handle.request_close_event, self._on_foreign_request_close)

            self.qtile.manage(self)

        if self.group and self.group.screen:
            self.mapped = True
            self.core.focus_window(self)

    def _on_unmap(self, _listener: Listener, _data: Any) -> None:
        logger.debug("Signal: xwindow unmap")
        self.mapped = False
        self.damage()
        seat = self.core.seat
        if not seat.destroyed:
            if self.surface.surface == seat.keyboard_state.focused_surface:
                seat.keyboard_clear_focus()

        if not self._unmapping:
            self.qtile.unmanage(self.wid)
            self.finalize()

        self._unmapping = False

    def _on_request_fullscreen(self, _listener: Listener, _data: Any) -> None:
        logger.debug("Signal: xwindow request_fullscreen")
        if self.qtile.config.auto_fullscreen:
            self.fullscreen = not self.fullscreen

    def _on_set_title(self, _listener: Listener, _data: Any) -> None:
        logger.debug("Signal: xwindow set_title")
        title = self.surface.title
        if title and title != self.name:
            self.name = title
            self.ftm_handle.set_title(title)
            hook.fire("client_name_updated", self)

    def _on_set_class(self, _listener: Listener, _data: Any) -> None:
        logger.debug("Signal: xwindow set_class")
        self._wm_class = self.surface.wm_class
        self.ftm_handle.set_app_id(self._wm_class or "")

    def hide(self) -> None:
        if self.mapped:
            self._unmapping = True
            self.surface.unmap_event.emit()

    def unhide(self) -> None:
        if not self.mapped:
            self.surface.map_event.emit()

    def kill(self) -> None:
        self.surface.close()

    def has_fixed_size(self) -> bool:
        hints = self.surface.size_hints
        # TODO: Maybe consider these flags too:
        # "PMinSize" in self.hints["flags"] and "PMaxSize" in self.hints["flags"]
        return bool(
            hints
            and 0 < hints.min_width == hints.max_width
            and 0 < hints.min_height == hints.max_height
        )

    def is_transient_for(self) -> base.WindowType | None:
        """What window is this window a transient window for?"""
        parent = self.surface.parent
        if parent:
            for win in self.qtile.windows_map.values():
                if isinstance(win, XWindow) and win.surface == parent:
                    return win
        return None

    def get_pid(self) -> int:
        return self.surface.pid

    def get_wm_type(self) -> str | None:
        wm_type = self.surface.window_type
        if wm_type:
            return self.core.xwayland_atoms[wm_type[0]]
        return None

    def get_wm_role(self) -> str | None:
        return self.surface.role

    def _update_fullscreen(self, do_full: bool) -> None:
        if do_full != (self._float_state == FloatStates.FULLSCREEN):
            self.surface.set_fullscreen(do_full)
            self.ftm_handle.set_fullscreen(do_full)

    @property
    def fullscreen(self) -> bool:
        return self._float_state == FloatStates.FULLSCREEN

    @fullscreen.setter
    def fullscreen(self, do_full: bool) -> None:
        self.surface.set_fullscreen(do_full)
        if do_full:
            screen = (self.group and self.group.screen) or self.qtile.find_closest_screen(
                self.x, self.y
            )
            bw = self.group.floating_layout.fullscreen_border_width if self.group else 0
            self._reconfigure_floating(
                screen.x,
                screen.y,
                screen.width - 2 * bw,
                screen.height - 2 * bw,
                new_float_state=FloatStates.FULLSCREEN,
            )
        elif self._float_state == FloatStates.FULLSCREEN:
            self.floating = False

        self.ftm_handle.set_fullscreen(do_full)

    def place(
        self,
        x: int,
        y: int,
        width: int,
        height: int,
        borderwidth: int,
        bordercolor: ColorsType | None,
        above: bool = False,
        margin: int | list[int] | None = None,
        respect_hints: bool = False,
    ) -> None:
        # Adjust the placement to account for layout margins, if there are any.
        if margin is not None:
            if isinstance(margin, int):
                margin = [margin] * 4
            x += margin[3]
            y += margin[0]
            width -= margin[1] + margin[3]
            height -= margin[0] + margin[2]

        if respect_hints:
            hints = self.surface.size_hints
            if hints:
                width = max(width, hints.min_width)
                height = max(height, hints.min_height)
                if hints.max_width > 0:
                    width = min(width, hints.max_width)
                if hints.max_height > 0:
                    height = min(height, hints.max_height)

        # save x and y float offset
        if self.group is not None and self.group.screen is not None:
            self.float_x = x - self.group.screen.x
            self.float_y = y - self.group.screen.y

        self.x = x
        self.y = y
        self._width = int(width)
        self._height = int(height)
        self.surface.configure(x, y, self._width, self._height)
        self.paint_borders(bordercolor, borderwidth)

        if above:
            self.cmd_bring_to_front()

        prev_outputs = self._outputs.copy()
        self._find_outputs()
        for output in self._outputs | prev_outputs:
            output.damage()

    def cmd_static(
        self,
        screen: int | None = None,
        x: int | None = None,
        y: int | None = None,
        width: int | None = None,
        height: int | None = None,
    ) -> None:
        Window.cmd_static(self, screen, x, y, width, height)
        hook.fire("client_managed", self.qtile.windows_map[self._wid])

    def _to_static(self) -> XStatic:
        return XStatic(self.core, self.qtile, self.surface, self.wid, self._idle_inhibitors_count)


class Static(typing.Generic[S], _Base, base.Static, HasListeners):
    def __init__(
        self,
        core: Core,
        qtile: Qtile,
        surface: S,
        wid: int,
        idle_inhibitor_count: int = 0,
    ):
        base.Static.__init__(self)
        self.core = core
        self.qtile = qtile
        self.surface = surface
        self.screen = qtile.current_screen
        self._wid = wid
        self._mapped: bool = False
        self.x = 0
        self.y = 0
        self._width = 0
        self._height = 0
        self.borderwidth: int = 0
        self.bordercolor: list[ffi.CData] = [_rgb((0, 0, 0, 1))]
        self.opacity: float = 1.0
        self._outputs: set[Output] = set()
        self._wm_class: str | None = None
        self._idle_inhibitors_count = idle_inhibitor_count

        if surface.data:
            self.ftm_handle = surface.data
            self.add_listener(self.ftm_handle.request_close_event, self._on_foreign_request_close)

    def finalize(self) -> None:
        self.finalize_listeners()

    @property
    def wid(self) -> int:
        return self._wid

    @property
    def mapped(self) -> bool:
        return self._mapped

    @mapped.setter
    def mapped(self, mapped: bool) -> None:
        if mapped == self._mapped:
            return
        self._mapped = mapped

        if mapped:
            self.core.mapped_windows.append(self)
        else:
            self.core.mapped_windows.remove(self)

        self.core.stack_windows()

    def _find_outputs(self) -> None:
        self._outputs = set(o for o in self.core.outputs if o.contains(self))

    def damage(self) -> None:
        for output in self._outputs:
            output.damage()

    def focus(self, warp: bool) -> None:
        self.core.focus_window(self)

        if warp and self.qtile.config.cursor_warp:
            self.core.warp_pointer(
                self.x + self.width / 2,
                self.y + self.height / 2,
            )

        hook.fire("client_focus", self)

    def paint_borders(self, color: ColorsType | None, width: int) -> None:
        if color:
            if isinstance(color, list):
                if len(color) > width:
                    color = color[:width]
                self.bordercolor = [_rgb(c) for c in color]
            else:
                self.bordercolor = [_rgb(color)]
        self.borderwidth = width

    def _on_map(self, _listener: Listener, _data: Any) -> None:
        logger.debug("Signal: static map")
        self.mapped = True
        self.focus(True)

    def _on_unmap(self, _listener: Listener, _data: Any) -> None:
        logger.debug("Signal: static unmap")
        self.mapped = False
        if self.surface.surface == self.core.seat.keyboard_state.focused_surface:  # type: ignore
            group = self.qtile.current_screen.group
            if group.current_window:
                group.focus(group.current_window, warp=self.qtile.config.cursor_warp)
            else:
                self.core.seat.keyboard_clear_focus()
        self.damage()

    def _on_destroy(self, _listener: Listener, _data: Any) -> None:
        logger.debug("Signal: static destroy")
        if self.mapped:
            logger.warning("Window destroyed before unmap event.")
            self.mapped = False

        if self in self.core.pending_windows:
            self.core.pending_windows.remove(self)
        else:
            self.qtile.unmanage(self.wid)

        self.finalize()

    def _on_commit(self, _listener: Listener, _data: Any) -> None:
        self.damage()

    def _on_foreign_request_close(self, _listener: Listener, _data: Any) -> None:
        logger.debug("Signal: foreign_toplevel_management static request_close")
        self.kill()

    def _on_inhibitor_destroy(self, listener: Listener, surface: Surface) -> None:
        # We don't have reference to the inhibitor, but it doesn't really
        # matter we only need to keep count of how many inhibitors there are
        self._idle_inhibitors_count -= 1
        listener.remove()
        if self._idle_inhibitors_count == 0:
            self.core.check_idle_inhibitor()

    def add_idle_inhibitor(
        self, surface: Surface, _x: int, _y: int, inhibitor: IdleInhibitorV1 | None
    ) -> None:
        if inhibitor and surface == inhibitor.surface:
            self._idle_inhibitors_count += 1
            inhibitor.data = self
            self.add_listener(inhibitor.destroy_event, self._on_inhibitor_destroy)
            if self._idle_inhibitors_count == 1:
                self.core.check_idle_inhibitor()

    @property
    def is_idle_inhibited(self) -> bool:
        return self._idle_inhibitors_count > 0

    def cmd_bring_to_front(self) -> None:
        if self.mapped:
            self.core.mapped_windows.remove(self)
            self.core.mapped_windows.append(self)
            self.core.stack_windows()
            self.damage()


class XdgStatic(Static[XdgSurface]):
    """A static window belonging to the XDG shell."""

    def __init__(
        self,
        core: Core,
        qtile: Qtile,
        surface: XdgSurface,
        wid: int,
        idle_inhibitor_count: int,
    ):
        Static.__init__(
            self, core, qtile, surface, wid, idle_inhibitor_count=idle_inhibitor_count
        )
        self.subsurfaces: list[SubSurface] = []
        self._find_outputs()

        if surface.toplevel.title:
            self.name = surface.toplevel.title
        self._wm_class = surface.toplevel.app_id

        self.add_listener(surface.map_event, self._on_map)
        self.add_listener(surface.unmap_event, self._on_unmap)
        self.add_listener(surface.destroy_event, self._on_destroy)
        self.add_listener(surface.surface.commit_event, self._on_commit)
        self.add_listener(surface.toplevel.set_title_event, self._on_set_title)
        self.add_listener(surface.toplevel.set_app_id_event, self._on_set_app_id)

    def finalize(self) -> None:
        Static.finalize(self)
        for subsurface in self.subsurfaces:
            subsurface.finalize()

        for pc in self.core.pointer_constraints.copy():
            if pc.window is self:
                pc.finalize()

    def kill(self) -> None:
        self.surface.send_close()

    def hide(self) -> None:
        if self.mapped:
            self.surface.unmap_event.emit()

    def unhide(self) -> None:
        if not self.mapped:
            self.surface.map_event.emit()

    def place(
        self,
        x: int,
        y: int,
        width: int,
        height: int,
        borderwidth: int,
        bordercolor: ColorsType | None,
        above: bool = False,
        margin: int | list[int] | None = None,
        respect_hints: bool = False,
    ) -> None:
        self.x = x
        self.y = y
        self._width = width
        self._height = height
        self.surface.set_size(self._width, self._height)
        self.paint_borders(bordercolor, borderwidth)
        self._find_outputs()
        self.damage()

    def _on_set_title(self, _listener: Listener, _data: Any) -> None:
        logger.debug("Signal: xdgstatic set_title")
        title = self.surface.toplevel.title
        if title and title != self.name:
            self.name = title
            self.ftm_handle.set_title(self.name)
            hook.fire("client_name_updated", self)

    def _on_set_app_id(self, _listener: Listener, _data: Any) -> None:
        logger.debug("Signal: xdgstatic set_app_id")
        self._wm_class = self.surface.toplevel.app_id
        self.ftm_handle.set_app_id(self._wm_class or "")


class XStatic(Static[xwayland.Surface]):
    """A static window belonging to the XWayland shell."""

    def __init__(
        self,
        core: Core,
        qtile: Qtile,
        surface: xwayland.Surface,
        wid: int,
        idle_inhibitor_count: int,
    ):
        Static.__init__(
            self, core, qtile, surface, wid, idle_inhibitor_count=idle_inhibitor_count
        )
        self._wm_class = surface.wm_class
        self._find_outputs()

        self.add_listener(surface.map_event, self._on_map)
        self.add_listener(surface.unmap_event, self._on_unmap)
        self.add_listener(surface.destroy_event, self._on_destroy)
        self.add_listener(surface.surface.commit_event, self._on_commit)
        self.add_listener(surface.set_title_event, self._on_set_title)
        self.add_listener(surface.set_class_event, self._on_set_class)

    def kill(self) -> None:
        self.surface.close()

    def hide(self) -> None:
        if self.mapped:
            self.surface.unmap_event.emit()

    def unhide(self) -> None:
        if not self.mapped:
            self.surface.map_event.emit()

    def place(
        self,
        x: int,
        y: int,
        width: int,
        height: int,
        borderwidth: int,
        bordercolor: ColorsType | None,
        above: bool = False,
        margin: int | list[int] | None = None,
        respect_hints: bool = False,
    ) -> None:
        self.x = x
        self.y = y
        self._width = width
        self._height = height
        self.surface.configure(x, y, self._width, self._height)
        self.paint_borders(bordercolor, borderwidth)
        self._find_outputs()
        self.damage()

    def _on_set_title(self, _listener: Listener, _data: Any) -> None:
        logger.debug("Signal: xstatic set_title")
        title = self.surface.title
        if title and title != self.name:
            self.name = title
            self.ftm_handle.set_title(title)
            hook.fire("client_name_updated", self)

    def _on_set_class(self, _listener: Listener, _data: Any) -> None:
        logger.debug("Signal: xstatic set_class")
        self._wm_class = self.surface.wm_class
        self.ftm_handle.set_app_id(self._wm_class or "")


class LayerStatic(Static[LayerSurfaceV1]):
    """A static window belonging to the layer shell."""

    def __init__(
        self,
        core: Core,
        qtile: Qtile,
        surface: LayerSurfaceV1,
        wid: int,
    ):
        Static.__init__(self, core, qtile, surface, wid)
        self.subsurfaces: list[SubSurface] = []

        self.add_listener(surface.map_event, self._on_map)
        self.add_listener(surface.unmap_event, self._on_unmap)
        self.add_listener(surface.destroy_event, self._on_destroy)
        self.add_listener(surface.surface.commit_event, self._on_commit)

        self._layer = LayerShellV1Layer.BACKGROUND
        self.desired_width = 0
        self.desired_height = 0
        if surface.output is None:
            surface.output = core.output_layout.output_at(core.cursor.x, core.cursor.y)

        if surface.output:
            output = surface.output.data
            if output:
                self.output = output
                self.screen = self.output.screen

        self.mapped = True
        self._outputs.add(self.output)

    def finalize(self) -> None:
        Static.finalize(self)
        for subsurface in self.subsurfaces:
            subsurface.finalize()

        for pc in self.core.pointer_constraints.copy():
            if pc.window is self:
                pc.finalize()

    @property
    def mapped(self) -> bool:
        return self._mapped

    @mapped.setter
    def mapped(self, mapped: bool) -> None:
        if mapped == self._mapped:
            return
        self._mapped = mapped

        self._layer = self.surface.pending.layer
        layer = self.output.layers[self._layer]
        if mapped:
            layer.append(self)
        else:
            layer.remove(self)

            if self.reserved_space:
                self.qtile.free_reserved_space(self.reserved_space, self.screen)
        self.output.organise_layers()

        self.core.stack_windows()

    def _on_map(self, _listener: Listener, _data: Any) -> None:
        logger.debug("Signal: layerstatic map")
        self.mapped = True
        self.output.organise_layers()
        self.focus(True)

    def _on_unmap(self, _listener: Listener, _data: Any) -> None:
        logger.debug("Signal: layerstatic unmap")
        self.mapped = False
        if self.surface.surface == self.core.seat.keyboard_state.focused_surface:
            group = self.qtile.current_screen.group
            if group.current_window:
                group.focus(group.current_window, warp=self.qtile.config.cursor_warp)
            else:
                self.core.seat.keyboard_clear_focus()
        self.output.organise_layers()
        self.damage()

    def _on_commit(self, _listener: Listener, _data: Any) -> None:
        current = self.surface.current
        if (
            self._layer != current.layer
            or self.desired_width != current.desired_width
            or self.desired_height != current.desired_height
        ):
            self.output.organise_layers()
        self.damage()

    def kill(self) -> None:
        self.surface.destroy()

    def hide(self) -> None:
        if self.mapped:
            self.surface.unmap_event.emit()

    def unhide(self) -> None:
        if not self.mapped:
            self.surface.map_event.emit()

    def place(
        self,
        x: int,
        y: int,
        width: int,
        height: int,
        borderwidth: int,
        bordercolor: ColorsType | None,
        above: bool = False,
        margin: int | list[int] | None = None,
        respect_hints: bool = False,
    ) -> None:
        self.x = x
        self.y = y
        self._width = int(width)
        self._height = int(height)
        self.surface.configure(self._width, self._height)
        self.damage()

    def cmd_bring_to_front(self) -> None:
        pass


class Internal(_Base, base.Internal):
    """
    Internal windows are simply textures controlled by the compositor.
    """

    def __init__(self, core: Core, qtile: Qtile, x: int, y: int, width: int, height: int):
        self.core = core
        self.qtile = qtile
        self._mapped: bool = False
        self._wid: int = self.core.new_wid()
        self.x: int = x
        self.y: int = y
        self._width: int = width
        self._height: int = height
        self._opacity: float = 1.0
        self._outputs: set[Output] = set(o for o in self.core.outputs if o.contains(self))
        self.texture: Texture = self._new_texture()

    def finalize(self) -> None:
        self.hide()

    def _new_texture(self) -> Texture:
        clear = cairocffi.ImageSurface(cairocffi.FORMAT_ARGB32, self.width, self.height)
        with cairocffi.Context(clear) as context:
            context.set_source_rgba(0, 0, 0, 0)
            context.paint()

        return Texture.from_pixels(
            self.core.renderer,
            DRM_FORMAT_ARGB8888,
            cairocffi.ImageSurface.format_stride_for_width(cairocffi.FORMAT_ARGB32, self.width),
            self.width,
            self.height,
            cairocffi.cairo.cairo_image_surface_get_data(clear._pointer),
        )

    def create_drawer(self, width: int, height: int) -> Drawer:
        """Create a Drawer that draws to this window."""
        return Drawer(self.qtile, self, width, height)

    @property
    def mapped(self) -> bool:
        return self._mapped

    @mapped.setter
    def mapped(self, mapped: bool) -> None:
        """We keep track of which windows are mapped to we know which to render"""
        if mapped == self._mapped:
            return
        self._mapped = mapped
        if mapped:
            self.core.mapped_windows.append(self)
        else:
            self.core.mapped_windows.remove(self)
        self.core.stack_windows()

    def hide(self) -> None:
        self.mapped = False
        self.damage()

    def unhide(self) -> None:
        self.mapped = True
        self.damage()

    def focus(self, warp: bool) -> None:
        self.core.focus_window(self)

    def kill(self) -> None:
        self.hide()
        if self.wid in self.qtile.windows_map:
            # It will be present during config reloads; absent during shutdown as this
            # will follow graceful_shutdown
            del self.qtile.windows_map[self.wid]

    def damage(self) -> None:
        for output in self._outputs:
            output.damage()

    def place(
        self,
        x: int,
        y: int,
        width: int,
        height: int,
        borderwidth: int,
        bordercolor: ColorsType | None,
        above: bool = False,
        margin: int | list[int] | None = None,
        respect_hints: bool = False,
    ) -> None:
        if above and self._mapped:
            self.core.mapped_windows.remove(self)
            self.core.mapped_windows.append(self)
            self.core.stack_windows()

        self.x = x
        self.y = y
        needs_reset = width != self.width or height != self.height
        self._width = width
        self._height = height

        if needs_reset:
            self.texture = self._new_texture()

        self._outputs = set(o for o in self.core.outputs if o.contains(self))
        self.damage()

    def info(self) -> dict:
        """Return a dictionary of info."""
        return dict(
            x=self.x,
            y=self.y,
            width=self.width,
            height=self.height,
            id=self.wid,
        )

    def cmd_bring_to_front(self) -> None:
        if self.mapped:
            self.core.mapped_windows.remove(self)
            self.core.mapped_windows.append(self)
            self.core.stack_windows()


WindowType = typing.Union[Window, Static, Internal]


class XdgPopupWindow(HasListeners):
    """
    This represents a single `struct wlr_xdg_popup` object and is owned by a single
    parent window (of `WindowType | XdgPopupWindow`). wlroots does most of the
    work for us, but we need to listen to certain events so that we know when to render
    frames and we need to unconstrain the popups so they are completely visible.
    """

    def __init__(self, parent: XdgWindow | XdgPopupWindow, xdg_popup: XdgPopup):
        self.parent = parent
        self.xdg_popup = xdg_popup
        self.core: Core = parent.core
        self.popups: list[XdgPopupWindow] = []

        # Keep on output
        if isinstance(parent, XdgPopupWindow):
            # This is a nested XdgPopup
            self.output: Output = parent.output
            self.output_box: Box = parent.output_box
        else:
            # Parent is an XdgSurface; This is a first-level XdgPopup
            box = xdg_popup.base.get_geometry()
            lx, ly = self.core.output_layout.closest_point(parent.x + box.x, parent.y + box.y)
            wlr_output = self.core.output_layout.output_at(lx, ly)
            assert wlr_output and isinstance(wlr_output.data, Output)
            self.output = wlr_output.data
            box = Box(*self.output.get_geometry())
            box.x = round(box.x - lx)
            box.y = round(box.y - ly)
            self.output_box = box
        xdg_popup.unconstrain_from_box(self.output_box)

        self.add_listener(xdg_popup.base.map_event, self._on_map)
        self.add_listener(xdg_popup.base.unmap_event, self._on_unmap)
        self.add_listener(xdg_popup.base.destroy_event, self._on_destroy)
        self.add_listener(xdg_popup.base.new_popup_event, self._on_new_popup)
        self.add_listener(xdg_popup.base.surface.commit_event, self._on_commit)

    def _on_map(self, _listener: Listener, _data: Any) -> None:
        logger.debug("Signal: popup map")
        self.output.damage()

    def _on_unmap(self, _listener: Listener, _data: Any) -> None:
        logger.debug("Signal: popup unmap")
        self.output.damage()

    def _on_destroy(self, _listener: Listener, _data: Any) -> None:
        logger.debug("Signal: popup destroy")
        self.finalize_listeners()
        self.output.damage()

    def _on_new_popup(self, _listener: Listener, xdg_popup: XdgPopup) -> None:
        logger.debug("Signal: popup new_popup")
        self.popups.append(XdgPopupWindow(self, xdg_popup))

    def _on_commit(self, _listener: Listener, _data: Any) -> None:
        self.output.damage()


class SubSurface(HasListeners):
    """
    This represents a single `struct wlr_subsurface` object and is owned by a single
    parent window (of `WindowType | SubSurface`). We only need to track them so
    that we can listen to their commit events and render accordingly.
    """

    def __init__(self, parent: XdgWindow | SubSurface, subsurface: WlrSubSurface):
        self.parent = parent
        self.subsurfaces: list[SubSurface] = []

        self.add_listener(subsurface.destroy_event, self._on_destroy)
        self.add_listener(subsurface.surface.commit_event, parent._on_commit)
        self.add_listener(subsurface.surface.new_subsurface_event, self._on_new_subsurface)

    def finalize(self) -> None:
        self.finalize_listeners()
        for subsurface in self.subsurfaces:
            subsurface.finalize()
        self.parent.subsurfaces.remove(self)

    def _on_destroy(self, _listener: Listener, _data: Any) -> None:
        self.finalize()

    def _on_commit(self, listener: Listener, _data: Any) -> None:
        self.parent._on_commit(listener, None)

    def _on_new_subsurface(self, _listener: Listener, subsurface: WlrSubSurface) -> None:
        self.subsurfaces.append(SubSurface(self, subsurface))


class PointerConstraint(HasListeners):
    """
    A small object to listen to signals on `struct wlr_pointer_constraint_v1` instances.
    """

    rect: Box

    def __init__(self, core: Core, wlr_constraint: PointerConstraintV1):
        self.core = core
        self.wlr_constraint = wlr_constraint
        self._warp_target: tuple[float, float] = (0, 0)
        self._needs_warp: bool = False

        self.add_listener(wlr_constraint.set_region_event, self._on_set_region)
        self.add_listener(wlr_constraint.destroy_event, self._on_destroy)

        owner = None

        if core.qtile and core.qtile.windows_map:
            for win in core.qtile.windows_map.values():
                if isinstance(win, (XdgWindow, XdgStatic)):
                    if win.surface.surface == self.wlr_constraint.surface:
                        owner = win
                        break

        if owner is None:
            logger.error("No window found for pointer constraints. Please report.")
            raise RuntimeError

        self.window: XdgWindow | XdgStatic = owner

    def finalize(self) -> None:
        if self.core.active_pointer_constraint is self:
            self.disable()
        self.finalize_listeners()
        self.core.pointer_constraints.remove(self)

    def _on_set_region(self, _listener: Listener, _data: Any) -> None:
        logger.debug("Signal: wlr_pointer_constraint_v1 set_region")
        self._get_region()

    def _on_destroy(self, _listener: Listener, wlr_constraint: PointerConstraintV1) -> None:
        logger.debug("Signal: wlr_pointer_constraint_v1 destroy")
        self.finalize()

    def _on_commit(self, _listener: Listener, _data: Any) -> None:
        if self._needs_warp:
            # Warp in case the pointer is not inside the rect
            if not self.rect.contains_point(self.core.cursor.x, self.core.cursor.y):
                self.core.warp_pointer(*self._warp_target)
            self._needs_warp = False

    def _get_region(self) -> None:
        rect = self.wlr_constraint.region.rectangles_as_boxes()[0]
        rect.x += self.window.x + self.window.borderwidth
        rect.y += self.window.y + self.window.borderwidth
        self._warp_target = (rect.x + rect.width / 2, rect.y + rect.height / 2)
        self.rect = rect
        self._needs_warp = True

    def enable(self) -> None:
        logger.debug("Enabling pointer constraints.")
        self.core.active_pointer_constraint = self
        self._get_region()
        self.add_listener(self.wlr_constraint.surface.commit_event, self._on_commit)
        self.wlr_constraint.send_activated()

    def disable(self) -> None:
        logger.debug("Disabling pointer constraints.")

        if self.wlr_constraint.current.committed & PointerConstraintV1StateField.CURSOR_HINT:
            x, y = self.wlr_constraint.current.cursor_hint
            self.core.warp_pointer(x + self.window.x, y + self.window.y)

        self.core.active_pointer_constraint = None
        self.wlr_constraint.send_deactivated()



# Copyright (c) 2011 Florian Mounier
# Copyright (c) 2012-2013 Craig Barnes
# Copyright (c) 2012 roger
# Copyright (c) 2012, 2014-2015 Tycho Andersen
# Copyright (c) 2014 Sean Vig
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import tempfile
from pathlib import Path

import pytest

import libqtile.bar
import libqtile.config
import libqtile.confreader
import libqtile.layout
import libqtile.widget
from libqtile.command.base import CommandError
from test.helpers import Retry


class GBConfig(libqtile.confreader.Config):
    auto_fullscreen = True
    keys = []
    mouse = []
    groups = [
        libqtile.config.Group("a"),
        libqtile.config.Group("bb"),
        libqtile.config.Group("ccc"),
        libqtile.config.Group("dddd"),
        libqtile.config.Group("Pppy"),
    ]
    layouts = [libqtile.layout.stack.Stack(num_stacks=1)]
    floating_layout = libqtile.resources.default_config.floating_layout
    screens = [
        libqtile.config.Screen(
            top=libqtile.bar.Bar(
                [
                    libqtile.widget.CPUGraph(
                        width=libqtile.bar.STRETCH,
                        type="linefill",
                        border_width=20,
                        margin_x=1,
                        margin_y=1,
                    ),
                    libqtile.widget.MemoryGraph(type="line"),
                    libqtile.widget.SwapGraph(type="box"),
                    libqtile.widget.TextBox(name="text", background="333333"),
                ],
                50,
            ),
            bottom=libqtile.bar.Bar(
                [
                    libqtile.widget.GroupBox(),
                    libqtile.widget.AGroupBox(),
                    libqtile.widget.Prompt(),
                    libqtile.widget.WindowName(),
                    libqtile.widget.Sep(),
                    libqtile.widget.Clock(),
                ],
                50,
            ),
            # TODO: Add vertical bars and test widgets that support them
        )
    ]


gb_config = pytest.mark.parametrize("manager", [GBConfig], indirect=True)


def test_completion():
    c = libqtile.widget.prompt.CommandCompleter(None, True)
    c.reset()
    c.lookup = [
        ("a", "x/a"),
        ("aa", "x/aa"),
    ]
    assert c.complete("a") == "a"
    assert c.actual() == "x/a"
    assert c.complete("a") == "aa"
    assert c.complete("a") == "a"

    c = libqtile.widget.prompt.CommandCompleter(None)
    r = c.complete("l")
    assert c.actual().endswith(r)

    c.reset()
    assert c.complete("/bi") == "/bin/"
    c.reset()
    assert c.complete("/bin") != "/bin/"
    c.reset()

    home_dir = os.path.expanduser("~")
    with tempfile.TemporaryDirectory(prefix="qtile_test_", dir=home_dir) as absolute_tmp_path:
        tmp_dirname = absolute_tmp_path[len(home_dir + os.sep) :]
        user_input = os.path.join("~", tmp_dirname)
        assert c.complete(user_input) == user_input

        c.reset()
        test_bin_dir = os.path.join(absolute_tmp_path, "qtile-test-bin")
        os.mkdir(test_bin_dir)
        assert c.complete(user_input) == os.path.join(user_input, "qtile-test-bin") + os.sep

    c.reset()
    s = "thisisatotallynonexistantpathforsure"
    assert c.complete(s) == s
    assert c.actual() == s
    c.reset()


@gb_config
def test_draw(manager):
    manager.test_window("one")
    b = manager.c.bar["bottom"].info()
    assert b["widgets"][0]["name"] == "groupbox"


@gb_config
def test_prompt(manager, monkeypatch):
    assert manager.c.widget["prompt"].info()["width"] == 0
    manager.c.spawncmd(":")
    manager.c.widget["prompt"].fake_keypress("a")
    manager.c.widget["prompt"].fake_keypress("Tab")

    manager.c.spawncmd(":")
    manager.c.widget["prompt"].fake_keypress("slash")
    manager.c.widget["prompt"].fake_keypress("Tab")

    script = Path(__file__).parent / "scripts" / "window.py"
    manager.c.spawncmd(":", aliases={"w": script.as_posix()})
    manager.c.widget["prompt"].fake_keypress("w")
    manager.c.widget["prompt"].fake_keypress("Return")

    @Retry(ignore_exceptions=(CommandError,))
    def is_spawned():
        return manager.c.window.info()

    is_spawned()


@gb_config
def test_event(manager):
    manager.c.group["bb"].toscreen()


@gb_config
def test_textbox(manager):
    assert "text" in manager.c.list_widgets()
    s = "some text"
    manager.c.widget["text"].update(s)
    assert manager.c.widget["text"].get() == s
    s = "Aye, much longer string than the initial one"
    manager.c.widget["text"].update(s)
    assert manager.c.widget["text"].get() == s
    manager.c.group["Pppy"].toscreen()
    manager.c.widget["text"].set_font(fontsize=12)


@gb_config
def test_textbox_errors(manager):
    manager.c.widget["text"].update(None)
    manager.c.widget["text"].update("".join(chr(i) for i in range(255)))
    manager.c.widget["text"].update("V\xE2r\xE2na\xE7\xEE")
    manager.c.widget["text"].update("\ua000")


@gb_config
def test_groupbox_button_press(manager):
    manager.c.group["ccc"].toscreen()
    assert manager.c.groups()["a"]["screen"] is None
    manager.c.bar["bottom"].fake_button_press(0, "bottom", 10, 10, 1)
    assert manager.c.groups()["a"]["screen"] == 0


class GeomConf(libqtile.confreader.Config):
    auto_fullscreen = False
    keys = []
    mouse = []
    groups = [
        libqtile.config.Group("a"),
        libqtile.config.Group("b"),
        libqtile.config.Group("c"),
        libqtile.config.Group("d"),
    ]
    layouts = [libqtile.layout.stack.Stack(num_stacks=1)]
    floating_layout = libqtile.resources.default_config.floating_layout
    screens = [
        libqtile.config.Screen(
            top=libqtile.bar.Bar([], 10),
            bottom=libqtile.bar.Bar([], 10),
            left=libqtile.bar.Bar([], 10),
            right=libqtile.bar.Bar([], 10),
        )
    ]


geom_config = pytest.mark.parametrize("manager", [GeomConf], indirect=True)


class DBarH(libqtile.bar.Bar):
    def __init__(self, widgets, size):
        libqtile.bar.Bar.__init__(self, widgets, size)
        self.horizontal = True


class DBarV(libqtile.bar.Bar):
    def __init__(self, widgets, size):
        libqtile.bar.Bar.__init__(self, widgets, size)
        self.horizontal = False


class DWidget:
    def __init__(self, length, length_type):
        self.length, self.length_type = length, length_type


@geom_config
def test_geometry(manager):
    manager.test_window("one")
    g = manager.c.screens()[0]["gaps"]
    assert g["top"] == (0, 0, 800, 10)
    assert g["bottom"] == (0, 590, 800, 10)
    assert g["left"] == (0, 10, 10, 580)
    assert g["right"] == (790, 10, 10, 580)
    assert len(manager.c.windows()) == 1
    geom = manager.c.windows()[0]
    assert geom["x"] == 10
    assert geom["y"] == 10
    assert geom["width"] == 778
    assert geom["height"] == 578
    internal = manager.c.internal_windows()
    assert len(internal) == 4


@geom_config
def test_resize(manager):
    def wd(dwidget_list):
        return [i.length for i in dwidget_list]

    def offx(dwidget_list):
        return [i.offsetx for i in dwidget_list]

    def offy(dwidget_list):
        return [i.offsety for i in dwidget_list]

    for DBar, off in ((DBarH, offx), (DBarV, offy)):  # noqa: N806
        b = DBar([], 100)

        dwidget_list = [
            DWidget(10, libqtile.bar.CALCULATED),
            DWidget(None, libqtile.bar.STRETCH),
            DWidget(None, libqtile.bar.STRETCH),
            DWidget(10, libqtile.bar.CALCULATED),
        ]
        b._resize(100, dwidget_list)
        assert wd(dwidget_list) == [10, 40, 40, 10]
        assert off(dwidget_list) == [0, 10, 50, 90]

        b._resize(101, dwidget_list)
        assert wd(dwidget_list) == [10, 41, 40, 10]
        assert off(dwidget_list) == [0, 10, 51, 91]

        dwidget_list = [DWidget(10, libqtile.bar.CALCULATED)]
        b._resize(100, dwidget_list)
        assert wd(dwidget_list) == [10]
        assert off(dwidget_list) == [0]

        dwidget_list = [DWidget(10, libqtile.bar.CALCULATED), DWidget(None, libqtile.bar.STRETCH)]
        b._resize(100, dwidget_list)
        assert wd(dwidget_list) == [10, 90]
        assert off(dwidget_list) == [0, 10]

        dwidget_list = [
            DWidget(None, libqtile.bar.STRETCH),
            DWidget(10, libqtile.bar.CALCULATED),
        ]
        b._resize(100, dwidget_list)
        assert wd(dwidget_list) == [90, 10]
        assert off(dwidget_list) == [0, 90]

        dwidget_list = [
            DWidget(10, libqtile.bar.CALCULATED),
            DWidget(None, libqtile.bar.STRETCH),
            DWidget(10, libqtile.bar.CALCULATED),
        ]
        b._resize(100, dwidget_list)
        assert wd(dwidget_list) == [10, 80, 10]
        assert off(dwidget_list) == [0, 10, 90]


class ExampleWidget(libqtile.widget.base._Widget):
    orientations = libqtile.widget.base.ORIENTATION_HORIZONTAL

    def __init__(self):
        libqtile.widget.base._Widget.__init__(self, 10)

    def draw(self):
        pass


class BrokenWidget(libqtile.widget.base._Widget):
    def __init__(self, exception_class, **config):
        libqtile.widget.base._Widget.__init__(self, 10, **config)
        self.exception_class = exception_class

    def _configure(self, qtile, bar):
        raise self.exception_class


def test_basic(manager_nospawn):
    config = GeomConf
    config.screens = [
        libqtile.config.Screen(
            bottom=libqtile.bar.Bar(
                [
                    ExampleWidget(),
                    libqtile.widget.Spacer(libqtile.bar.STRETCH),
                    ExampleWidget(),
                    libqtile.widget.Spacer(libqtile.bar.STRETCH),
                    ExampleWidget(),
                    libqtile.widget.Spacer(libqtile.bar.STRETCH),
                    ExampleWidget(),
                ],
                10,
            )
        )
    ]

    manager_nospawn.start(config)

    i = manager_nospawn.c.bar["bottom"].info()
    assert i["widgets"][0]["offset"] == 0
    assert i["widgets"][1]["offset"] == 10
    assert i["widgets"][1]["width"] == 252
    assert i["widgets"][2]["offset"] == 262
    assert i["widgets"][3]["offset"] == 272
    assert i["widgets"][3]["width"] == 256
    assert i["widgets"][4]["offset"] == 528
    assert i["widgets"][5]["offset"] == 538
    assert i["widgets"][5]["width"] == 252
    assert i["widgets"][6]["offset"] == 790
    libqtile.hook.clear()


def test_singlespacer(manager_nospawn):
    config = GeomConf
    config.screens = [
        libqtile.config.Screen(
            bottom=libqtile.bar.Bar(
                [
                    libqtile.widget.Spacer(libqtile.bar.STRETCH),
                ],
                10,
            )
        )
    ]

    manager_nospawn.start(config)

    i = manager_nospawn.c.bar["bottom"].info()
    assert i["widgets"][0]["offset"] == 0
    assert i["widgets"][0]["width"] == 800
    libqtile.hook.clear()


def test_nospacer(manager_nospawn):
    config = GeomConf
    config.screens = [
        libqtile.config.Screen(bottom=libqtile.bar.Bar([ExampleWidget(), ExampleWidget()], 10))
    ]

    manager_nospawn.start(config)

    i = manager_nospawn.c.bar["bottom"].info()
    assert i["widgets"][0]["offset"] == 0
    assert i["widgets"][1]["offset"] == 10
    libqtile.hook.clear()


def test_consecutive_spacer(manager_nospawn):
    config = GeomConf
    config.screens = [
        libqtile.config.Screen(
            bottom=libqtile.bar.Bar(
                [
                    ExampleWidget(),  # Left
                    libqtile.widget.Spacer(libqtile.bar.STRETCH),
                    libqtile.widget.Spacer(libqtile.bar.STRETCH),
                    ExampleWidget(),  # Centre
                    ExampleWidget(),
                    libqtile.widget.Spacer(libqtile.bar.STRETCH),
                    ExampleWidget(),  # Right
                ],
                10,
            )
        )
    ]

    manager_nospawn.start(config)

    i = manager_nospawn.c.bar["bottom"].info()
    assert i["widgets"][0]["offset"] == 0
    assert i["widgets"][0]["width"] == 10
    assert i["widgets"][1]["offset"] == 10
    assert i["widgets"][1]["width"] == 190
    assert i["widgets"][2]["offset"] == 200
    assert i["widgets"][2]["width"] == 190
    assert i["widgets"][3]["offset"] == 390
    assert i["widgets"][3]["width"] == 10
    assert i["widgets"][4]["offset"] == 400
    assert i["widgets"][4]["width"] == 10
    assert i["widgets"][5]["offset"] == 410
    assert i["widgets"][5]["width"] == 380
    assert i["widgets"][6]["offset"] == 790
    assert i["widgets"][6]["width"] == 10
    libqtile.hook.clear()


def test_configure_broken_widgets(manager_nospawn):
    config = GeomConf

    widget_list = [
        BrokenWidget(ValueError),
        BrokenWidget(IndexError),
        BrokenWidget(IndentationError),
        BrokenWidget(TypeError),
        BrokenWidget(NameError),
        BrokenWidget(ImportError),
        libqtile.widget.Spacer(libqtile.bar.STRETCH),
    ]

    config.screens = [libqtile.config.Screen(bottom=libqtile.bar.Bar(widget_list, 10))]

    manager_nospawn.start(config)

    i = manager_nospawn.c.bar["bottom"].info()

    # Check that we have same number of widgets
    assert len(i["widgets"]) == len(widget_list)

    # Check each broken widget is replaced
    for index, widget in enumerate(widget_list):
        if isinstance(widget, BrokenWidget):
            assert i["widgets"][index]["name"] == "configerrorwidget"


def test_bar_hide_show_with_margin(manager_nospawn):
    """Check :
        - the height of a horizontal bar with its margins,
        - the ordinate of a unique window.
    after 3 successive actions :
        - creation
        - hidding the bar
        - unhidding the bar
    """
    config = GeomConf

    config.screens = [libqtile.config.Screen(top=libqtile.bar.Bar([], 12, margin=[5, 5, 5, 5]))]

    manager_nospawn.start(config)
    manager_nospawn.test_window("w")

    assert manager_nospawn.c.bar["top"].info().get("size") == 22
    assert manager_nospawn.c.windows()[0]["y"] == 22

    manager_nospawn.c.hide_show_bar("top")
    assert manager_nospawn.c.bar["top"].info().get("size") == 0
    assert manager_nospawn.c.windows()[0]["y"] == 0

    manager_nospawn.c.hide_show_bar("top")
    assert manager_nospawn.c.bar["top"].info().get("size") == 22
    assert manager_nospawn.c.windows()[0]["y"] == 22


def test_bar_border_horizontal(manager_nospawn):
    config = GeomConf

    config.screens = [
        libqtile.config.Screen(
            top=libqtile.bar.Bar(
                [libqtile.widget.Spacer()],
                12,
                margin=5,
                border_width=5,
            ),
            bottom=libqtile.bar.Bar(
                [libqtile.widget.Spacer()],
                12,
                margin=5,
                border_width=0,
            ),
        )
    ]

    manager_nospawn.start(config)

    top_info = manager_nospawn.c.bar["top"].info
    bottom_info = manager_nospawn.c.bar["bottom"].info

    # Screen is 800px wide so:
    # -top bar should have width of 800 - 5 - 5 - 5 - 5 = 780 (margin and border)
    # -bottom bar should have width of 800 - 5 - 5 = 790 (margin and no border)

    assert top_info()["width"] == 780
    assert bottom_info()["width"] == 790

    # Bar "height" should still be the value set in the config but "size" is
    # adjusted for margin and border:
    # -top bar should have size of 12 + 5 + 5 + 5 + 5 = 32 (margin and border)
    # -bottom bar should have size of 12 + 5 + 5 = 22 (margin and border)

    assert top_info()["height"] == 12
    assert top_info()["size"] == 32
    assert bottom_info()["height"] == 12
    assert bottom_info()["size"] == 22

    # Test widget offsets
    # Where there is a border, widget should be offset by that amount

    _, xoffset = manager_nospawn.c.bar["top"].eval("self.widgets[0].offsetx")
    assert xoffset == "5"

    _, yoffset = manager_nospawn.c.bar["top"].eval("self.widgets[0].offsety")
    assert xoffset == "5"

    # Where there is no border, this should be 0
    _, xoffset = manager_nospawn.c.bar["bottom"].eval("self.widgets[0].offsetx")
    assert xoffset == "0"

    _, yoffset = manager_nospawn.c.bar["bottom"].eval("self.widgets[0].offsety")
    assert xoffset == "0"


def test_bar_border_vertical(manager_nospawn):
    config = GeomConf

    config.screens = [
        libqtile.config.Screen(
            left=libqtile.bar.Bar(
                [libqtile.widget.Spacer()],
                12,
                margin=5,
                border_width=5,
            ),
            right=libqtile.bar.Bar(
                [libqtile.widget.Spacer()],
                12,
                margin=5,
                border_width=0,
            ),
        )
    ]

    manager_nospawn.start(config)

    left_info = manager_nospawn.c.bar["left"].info
    right_info = manager_nospawn.c.bar["right"].info

    # Screen is 600px tall so:
    # -left bar should have height of 600 - 5 - 5 - 5 - 5 = 580 (margin and border)
    # -right bar should have height of 600 - 5 - 5 = 590 (margin and no border)

    assert left_info()["height"] == 580
    assert right_info()["height"] == 590

    # Bar "width" should still be the value set in the config but "size" is
    # adjusted for margin and border:
    # -left bar should have size of 12 + 5 + 5 + 5 + 5 = 32 (margin and border)
    # -right bar should have size of 12 + 5 + 5 = 22 (margin and border)

    assert left_info()["width"] == 12
    assert left_info()["size"] == 32
    assert right_info()["width"] == 12
    assert right_info()["size"] == 22

    # Test widget offsets
    # Where there is a border, widget should be offset by that amount

    _, xoffset = manager_nospawn.c.bar["left"].eval("self.widgets[0].offsetx")
    assert xoffset == "5"

    _, yoffset = manager_nospawn.c.bar["left"].eval("self.widgets[0].offsety")
    assert xoffset == "5"

    # Where there is no border, this should be 0
    _, xoffset = manager_nospawn.c.bar["right"].eval("self.widgets[0].offsetx")
    assert xoffset == "0"

    _, yoffset = manager_nospawn.c.bar["right"].eval("self.widgets[0].offsety")
    assert xoffset == "0"



# Copyright (c) 2021 elParaguayo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Widget specific tests

import subprocess

import pytest

import libqtile.config
from libqtile.widget import cmus
from test.widgets.conftest import FakeBar


class MockCmusRemoteProcess:
    CalledProcessError = None
    EXTRA = [
        "set aaa_mode all",
        "set continue true",
        "set play_library true",
        "set play_sorted false",
        "set replaygain disabled",
        "set replaygain_limit true",
        "set replaygain_preamp 0.000000",
        "set repeat false",
        "set repeat_current false",
        "set shuffle false",
        "set softvol false",
        "set vol_left 100",
        "set vol_right 100",
    ]

    info = {}
    is_error = False
    index = 0

    @classmethod
    def reset(cls):
        cls.info = [
            [
                "status playing",
                "file /playing/file/rickroll.mp3",
                "duration 222",
                "position 14",
                "tag artist Rick Astley",
                "tag album Whenever You Need Somebody",
                "tag title Never Gonna Give You Up",
            ],
            [
                "status playing",
                "file http://playing/file/sweetcaroline.mp3",
                "duration 222",
                "position 14",
                "tag artist Neil Diamond",
                "tag album Greatest Hits",
                "tag title Sweet Caroline",
            ],
            [
                "status stopped",
                "file http://streaming.source/tomjones.m3u",
                "duration -1",
                "position -9",
                "tag title It's Not Unusual",
                "stream tomjones",
            ],
            [
                "status playing",
                "file /playing/file/always.mp3",
                "duration 222",
                "position 14",
                "tag artist Above & Beyond",
                "tag album Anjunabeats 14",
                "tag title Always - Tinlicker Extended Mix",
            ],
        ]
        cls.index = 0
        cls.is_error = False

    @classmethod
    def call_process(cls, cmd):
        if cls.is_error:
            raise subprocess.CalledProcessError(-1, cmd=cmd, output="Couldn't connect to cmus.")

        if cmd[1:] == ["-C", "status"]:
            track = cls.info[cls.index]
            track.extend(cls.EXTRA)
            output = "\n".join(track)
            return output

        elif cmd[1] == "-p":
            cls.info[cls.index][0] = "status playing"

        elif cmd[1] == "-u":
            if cls.info[cls.index][0] == "status playing":
                cls.info[cls.index][0] = "status paused"

            elif cls.info[cls.index][0] == "status paused":
                cls.info[cls.index][0] = "status playing"

        elif cmd[1] == "-n":
            cls.index = (cls.index + 1) % len(cls.info)

        elif cmd[1] == "-r":
            cls.index = (cls.index - 1) % len(cls.info)

    @classmethod
    def Popen(cls, cmd):  # noqa: N802
        cls.call_process(cmd)


def no_op(*args, **kwargs):
    pass


@pytest.fixture
def patched_cmus(monkeypatch):
    MockCmusRemoteProcess.reset()
    monkeypatch.setattr("libqtile.widget.cmus.subprocess", MockCmusRemoteProcess)
    monkeypatch.setattr(
        "libqtile.widget.cmus.subprocess.CalledProcessError", subprocess.CalledProcessError
    )
    monkeypatch.setattr(
        "libqtile.widget.cmus.base.ThreadPoolText.call_process",
        MockCmusRemoteProcess.call_process,
    )
    return cmus


def test_cmus(fake_qtile, patched_cmus, fake_window):
    widget = patched_cmus.Cmus()
    fakebar = FakeBar([widget], window=fake_window)
    widget._configure(fake_qtile, fakebar)
    text = widget.poll()
    assert text == "♫ Rick Astley - Never Gonna Give You Up"
    assert widget.layout.colour == widget.play_color

    widget.play()
    text = widget.poll()
    assert text == "♫ Rick Astley - Never Gonna Give You Up"
    assert widget.layout.colour == widget.noplay_color


def test_cmus_play_stopped(fake_qtile, patched_cmus, fake_window):
    widget = patched_cmus.Cmus()

    # Set track to a stopped item
    MockCmusRemoteProcess.index = 2
    fakebar = FakeBar([widget], window=fake_window)
    widget._configure(fake_qtile, fakebar)
    text = widget.poll()

    # It's stopped so colour should reflect this
    assert text == "♫ tomjones"
    assert widget.layout.colour == widget.noplay_color

    widget.play()
    text = widget.poll()
    assert text == "♫ tomjones"
    assert widget.layout.colour == widget.play_color


def test_cmus_buttons(minimal_conf_noscreen, manager_nospawn, patched_cmus):
    widget = patched_cmus.Cmus(update_interval=30)
    config = minimal_conf_noscreen
    config.screens = [libqtile.config.Screen(top=libqtile.bar.Bar([widget], 10))]
    manager_nospawn.start(config)
    topbar = manager_nospawn.c.bar["top"]

    cmuswidget = manager_nospawn.c.widget["cmus"]
    assert cmuswidget.info()["text"] == "♫ Rick Astley - Never Gonna Give You Up"

    # Play next track
    # Non-local file source so widget just displays title
    topbar.fake_button_press(0, "top", 0, 0, button=4)
    cmuswidget.eval("self.update(self.poll())")
    assert cmuswidget.info()["text"] == "♫ Sweet Caroline"

    # Play next track
    # Stream source so widget just displays stream info
    topbar.fake_button_press(0, "top", 0, 0, button=4)
    cmuswidget.eval("self.update(self.poll())")
    assert cmuswidget.info()["text"] == "♫ tomjones"

    # Play previous track
    # Non-local file source so widget just displays title
    topbar.fake_button_press(0, "top", 0, 0, button=5)
    cmuswidget.eval("self.update(self.poll())")
    assert cmuswidget.info()["text"] == "♫ Sweet Caroline"


def test_cmus_error_handling(fake_qtile, patched_cmus, fake_window):
    widget = patched_cmus.Cmus()
    MockCmusRemoteProcess.is_error = True
    fakebar = FakeBar([widget], window=fake_window)
    widget._configure(fake_qtile, fakebar)
    text = widget.poll()

    # Widget does nothing with error message so text is blank
    # TODO: update widget to show error?
    assert text == ""


def test_escape_text(fake_qtile, patched_cmus, fake_window):
    widget = patched_cmus.Cmus()

    # Set track to a stopped item
    MockCmusRemoteProcess.index = 3
    fakebar = FakeBar([widget], window=fake_window)
    widget._configure(fake_qtile, fakebar)
    text = widget.poll()

    # It's stopped so colour should reflect this
    assert text == "♫ Above &amp; Beyond - Always - Tinlicker Extended Mix"



# Copyright (c) 2021 elParaguayo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Widget specific tests

import sys
from importlib import reload
from types import ModuleType

import pytest

import libqtile.bar
import libqtile.config


def no_op(*args, **kwargs):
    pass


class FakePsutil(ModuleType):
    class virtual_memory:  # noqa: N801
        def __init__(self):
            self.used = 2534260736  # 2,474,864K, 2,417M, 2.36G
            self.total = 8180686848  # 7,988,952K, 7,802M, 7.72G
            self.free = 2354114560
            self.percent = 39.4
            self.buffers = 346394624
            self.active = 1132359680
            self.inactive = 3862183936
            self.shared = 516395008

    class swap_memory:  # noqa: N801
        def __init__(self):
            self.total = 8429498368
            self.used = 0
            self.free = 8429498368
            self.percent = 0.0


@pytest.fixture()
def patched_memory(
    monkeypatch,
):
    monkeypatch.setitem(sys.modules, "psutil", FakePsutil("psutil"))
    from libqtile.widget import memory

    reload(memory)
    return memory


def test_memory_defaults(manager_nospawn, minimal_conf_noscreen, patched_memory):
    """Test no text when free space over threshold"""
    widget = patched_memory.Memory()
    config = minimal_conf_noscreen
    config.screens = [libqtile.config.Screen(top=libqtile.bar.Bar([widget], 10))]
    manager_nospawn.start(config)
    assert manager_nospawn.c.widget["memory"].info()["text"] == " 2417M/ 7802M"


@pytest.mark.parametrize(
    "unit,expects",
    [
        ("G", " 2G/ 8G"),
        ("M", " 2417M/ 7802M"),
        ("K", " 2474864K/ 7988952K"),
        ("B", " 2534260736B/ 8180686848B"),
    ],
)
def test_memory_units(manager_nospawn, minimal_conf_noscreen, patched_memory, unit, expects):
    """Test no text when free space over threshold"""
    widget = patched_memory.Memory(measure_mem=unit)
    config = minimal_conf_noscreen
    config.screens = [libqtile.config.Screen(top=libqtile.bar.Bar([widget], 10))]
    manager_nospawn.start(config)
    assert manager_nospawn.c.widget["memory"].info()["text"] == expects



# Copyright (c) 2015 Sean Vig
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pytest

import libqtile.config
from libqtile import layout
from libqtile.confreader import Config
from test.layouts.layout_utils import assert_dimensions, assert_focus_path, assert_focused


class MonadTallConfig(Config):
    auto_fullscreen = True
    groups = [libqtile.config.Group("a")]
    layouts = [layout.MonadTall()]
    floating_layout = libqtile.resources.default_config.floating_layout
    keys = []
    mouse = []
    screens = []
    follow_mouse_focus = False


monadtall_config = pytest.mark.parametrize("manager", [MonadTallConfig], indirect=True)


class MonadTallNCPBeforeCurrentConfig(Config):
    auto_fullscreen = True
    groups = [libqtile.config.Group("a")]
    layouts = [layout.MonadTall(new_client_position="before_current")]
    floating_layout = libqtile.resources.default_config.floating_layout
    keys = []
    mouse = []
    screens = []
    follow_mouse_focus = False


monadtallncpbeforecurrent_config = pytest.mark.parametrize(
    "manager", [MonadTallNCPBeforeCurrentConfig], indirect=True
)


class MonadTallNCPAfterCurrentConfig(Config):
    auto_fullscreen = True
    groups = [libqtile.config.Group("a")]
    layouts = [layout.MonadTall(new_client_position="after_current")]
    floating_layout = libqtile.resources.default_config.floating_layout
    keys = []
    mouse = []
    screens = []
    follow_mouse_focus = False


monadtallncpaftercurrent_config = pytest.mark.parametrize(
    "manager", [MonadTallNCPAfterCurrentConfig], indirect=True
)


class MonadTallNewCLientPositionBottomConfig(Config):
    auto_fullscreen = True
    groups = [libqtile.config.Group("a")]
    layouts = [layout.MonadTall(new_client_position="bottom")]
    floating_layout = libqtile.resources.default_config.floating_layout
    keys = []
    mouse = []
    screens = []
    follow_mouse_focus = False


class MonadTallMarginsConfig(Config):
    auto_fullscreen = True
    groups = [libqtile.config.Group("a")]
    layouts = [layout.MonadTall(margin=4)]
    floating_layout = libqtile.resources.default_config.floating_layout
    keys = []
    mouse = []
    screens = []
    follow_mouse_focus = False


monadtallmargins_config = pytest.mark.parametrize(
    "manager", [MonadTallMarginsConfig], indirect=True
)


class MonadWideConfig(Config):
    auto_fullscreen = True
    groups = [libqtile.config.Group("a")]
    layouts = [layout.MonadWide()]
    floating_layout = libqtile.resources.default_config.floating_layout
    keys = []
    mouse = []
    screens = []
    follow_mouse_focus = False


monadwide_config = pytest.mark.parametrize("manager", [MonadWideConfig], indirect=True)


class MonadWideNewClientPositionTopConfig(Config):
    auto_fullscreen = True
    groups = [libqtile.config.Group("a")]
    layouts = [layout.MonadWide(new_client_position="top")]
    floating_layout = libqtile.resources.default_config.floating_layout
    keys = []
    mouse = []
    screens = []
    follow_mouse_focus = False


class MonadWideMarginsConfig(Config):
    auto_fullscreen = True
    groups = [libqtile.config.Group("a")]
    layouts = [layout.MonadWide(margin=4)]
    floating_layout = libqtile.resources.default_config.floating_layout
    keys = []
    mouse = []
    screens = []
    follow_mouse_focus = False


@monadtall_config
def test_tall_add_clients(manager):
    manager.test_window("one")
    manager.test_window("two")
    assert manager.c.layout.info()["main"] == "one"
    assert manager.c.layout.info()["secondary"] == ["two"]
    assert_focused(manager, "two")

    manager.test_window("three")
    assert manager.c.layout.info()["main"] == "one"
    assert manager.c.layout.info()["secondary"] == ["two", "three"]
    assert_focused(manager, "three")

    manager.c.layout.previous()
    assert_focused(manager, "two")

    manager.test_window("four")
    assert manager.c.layout.info()["main"] == "one"
    assert manager.c.layout.info()["secondary"] == ["two", "four", "three"]
    assert_focused(manager, "four")


@monadtallncpbeforecurrent_config
def test_tall_add_clients_before_current(manager):
    """Test add client with new_client_position = before_current."""
    manager.test_window("one")
    manager.test_window("two")
    manager.test_window("three")
    assert manager.c.layout.info()["main"] == "three"
    assert manager.c.layout.info()["secondary"] == ["two", "one"]
    manager.c.layout.next()
    assert_focused(manager, "two")
    manager.test_window("four")
    assert manager.c.layout.info()["main"] == "three"
    assert manager.c.layout.info()["secondary"] == ["four", "two", "one"]
    assert_focused(manager, "four")


@monadtallncpaftercurrent_config
def test_tall_add_clients_after_current(manager):
    manager.test_window("one")
    manager.test_window("two")
    manager.test_window("three")
    manager.c.layout.previous()
    assert_focused(manager, "two")
    manager.test_window("four")
    assert manager.c.layout.info()["main"] == "one"
    assert manager.c.layout.info()["secondary"] == ["two", "four", "three"]
    assert_focused(manager, "four")


@pytest.mark.parametrize("manager", [MonadTallNewCLientPositionBottomConfig], indirect=True)
def test_tall_add_clients_at_bottom(manager):
    manager.test_window("one")
    manager.test_window("two")
    manager.test_window("three")
    manager.c.layout.previous()
    assert_focused(manager, "two")
    manager.test_window("four")
    assert manager.c.layout.info()["main"] == "one"
    assert manager.c.layout.info()["secondary"] == ["two", "three", "four"]


@monadwide_config
def test_wide_add_clients(manager):
    manager.test_window("one")
    manager.test_window("two")
    assert manager.c.layout.info()["main"] == "one"
    assert manager.c.layout.info()["secondary"] == ["two"]
    assert_focused(manager, "two")

    manager.test_window("three")
    assert manager.c.layout.info()["main"] == "one"
    assert manager.c.layout.info()["secondary"] == ["two", "three"]
    assert_focused(manager, "three")

    manager.c.layout.previous()
    assert_focused(manager, "two")

    manager.test_window("four")
    assert manager.c.layout.info()["main"] == "one"
    assert manager.c.layout.info()["secondary"] == ["two", "four", "three"]
    assert_focused(manager, "four")


@pytest.mark.parametrize("manager", [MonadWideNewClientPositionTopConfig], indirect=True)
def test_wide_add_clients_new_client_postion_top(manager):
    manager.test_window("one")
    manager.test_window("two")
    assert manager.c.layout.info()["main"] == "two"
    assert manager.c.layout.info()["secondary"] == ["one"]
    assert_focused(manager, "two")

    manager.test_window("three")
    assert manager.c.layout.info()["main"] == "three"
    assert manager.c.layout.info()["secondary"] == ["two", "one"]
    assert_focused(manager, "three")

    manager.c.layout.next()
    assert_focused(manager, "two")

    manager.test_window("four")
    assert manager.c.layout.info()["main"] == "four"
    assert manager.c.layout.info()["secondary"] == ["three", "two", "one"]
    assert_focused(manager, "four")


@monadtallmargins_config
def test_tall_margins(manager):
    manager.test_window("one")
    assert_dimensions(manager, 4, 4, 788, 588)

    manager.test_window("two")
    assert_focused(manager, "two")
    assert_dimensions(manager, 404, 4, 388, 588)

    manager.c.layout.previous()
    assert_focused(manager, "one")
    assert_dimensions(manager, 4, 4, 392, 588)


@pytest.mark.parametrize("manager", [MonadWideMarginsConfig], indirect=True)
def test_wide_margins(manager):
    manager.test_window("one")
    assert_dimensions(manager, 4, 4, 788, 588)

    manager.test_window("two")
    assert_focused(manager, "two")
    assert_dimensions(manager, 4, 304, 788, 288)

    manager.c.layout.previous()
    assert_focused(manager, "one")
    assert_dimensions(manager, 4, 4, 788, 292)


@monadtall_config
def test_tall_growmain_solosecondary(manager):
    manager.test_window("one")
    assert_dimensions(manager, 0, 0, 796, 596)

    manager.test_window("two")
    manager.c.layout.previous()
    assert_focused(manager, "one")

    assert_dimensions(manager, 0, 0, 396, 596)
    manager.c.layout.grow()
    # Grows 5% of 800 = 40 pixels
    assert_dimensions(manager, 0, 0, 436, 596)
    manager.c.layout.shrink()
    assert_dimensions(manager, 0, 0, 396, 596)

    # Max width is 75% of 800 = 600 pixels
    for _ in range(10):
        manager.c.layout.grow()
    assert_dimensions(manager, 0, 0, 596, 596)

    # Min width is 25% of 800 = 200 pixels
    for _ in range(10):
        manager.c.layout.shrink()
    assert_dimensions(manager, 0, 0, 196, 596)


@monadwide_config
def test_wide_growmain_solosecondary(manager):
    manager.test_window("one")
    assert_dimensions(manager, 0, 0, 796, 596)

    manager.test_window("two")
    manager.c.layout.previous()
    assert_focused(manager, "one")

    assert_dimensions(manager, 0, 0, 796, 296)
    manager.c.layout.grow()
    # Grows 5% of 800 = 30 pixels
    assert_dimensions(manager, 0, 0, 796, 326)
    manager.c.layout.shrink()
    assert_dimensions(manager, 0, 0, 796, 296)

    # Max width is 75% of 600 = 450 pixels
    for _ in range(10):
        manager.c.layout.grow()
    assert_dimensions(manager, 0, 0, 796, 446)

    # Min width is 25% of 600 = 150 pixels
    for _ in range(10):
        manager.c.layout.shrink()
    assert_dimensions(manager, 0, 0, 796, 146)


@monadtall_config
def test_tall_growmain_multiplesecondary(manager):
    manager.test_window("one")
    assert_dimensions(manager, 0, 0, 796, 596)

    manager.test_window("two")
    manager.test_window("three")
    manager.c.layout.previous()
    manager.c.layout.previous()
    assert_focused(manager, "one")

    assert_dimensions(manager, 0, 0, 396, 596)
    manager.c.layout.grow()
    # Grows 5% of 800 = 40 pixels
    assert_dimensions(manager, 0, 0, 436, 596)
    manager.c.layout.shrink()
    assert_dimensions(manager, 0, 0, 396, 596)

    # Max width is 75% of 800 = 600 pixels
    for _ in range(10):
        manager.c.layout.grow()
    assert_dimensions(manager, 0, 0, 596, 596)

    # Min width is 25% of 800 = 200 pixels
    for _ in range(10):
        manager.c.layout.shrink()
    assert_dimensions(manager, 0, 0, 196, 596)


@monadwide_config
def test_wide_growmain_multiplesecondary(manager):
    manager.test_window("one")
    assert_dimensions(manager, 0, 0, 796, 596)

    manager.test_window("two")
    manager.test_window("three")
    manager.c.layout.previous()
    manager.c.layout.previous()
    assert_focused(manager, "one")

    assert_dimensions(manager, 0, 0, 796, 296)
    manager.c.layout.grow()
    # Grows 5% of 600 = 30 pixels
    assert_dimensions(manager, 0, 0, 796, 326)
    manager.c.layout.shrink()
    assert_dimensions(manager, 0, 0, 796, 296)

    # Max width is 75% of 600 = 450 pixels
    for _ in range(10):
        manager.c.layout.grow()
    assert_dimensions(manager, 0, 0, 796, 446)

    # Min width is 25% of 600 = 150 pixels
    for _ in range(10):
        manager.c.layout.shrink()
    assert_dimensions(manager, 0, 0, 796, 146)


@monadtall_config
def test_tall_growsecondary_solosecondary(manager):
    manager.test_window("one")
    assert_dimensions(manager, 0, 0, 796, 596)

    manager.test_window("two")
    assert_focused(manager, "two")

    assert_dimensions(manager, 400, 0, 396, 596)
    manager.c.layout.grow()
    # Grows 5% of 800 = 40 pixels
    assert_dimensions(manager, 360, 0, 436, 596)
    manager.c.layout.shrink()
    assert_dimensions(manager, 400, 0, 396, 596)

    # Max width is 75% of 800 = 600 pixels
    for _ in range(10):
        manager.c.layout.grow()
    assert_dimensions(manager, 200, 0, 596, 596)

    # Min width is 25% of 800 = 200 pixels
    for _ in range(10):
        manager.c.layout.shrink()
    assert_dimensions(manager, 600, 0, 196, 596)


@monadwide_config
def test_wide_growsecondary_solosecondary(manager):
    manager.test_window("one")
    assert_dimensions(manager, 0, 0, 796, 596)

    manager.test_window("two")
    assert_focused(manager, "two")

    assert_dimensions(manager, 0, 300, 796, 296)
    manager.c.layout.grow()
    # Grows 5% of 600 = 30 pixels
    assert_dimensions(manager, 0, 270, 796, 326)
    manager.c.layout.shrink()
    assert_dimensions(manager, 0, 300, 796, 296)

    # Max width is 75% of 600 = 450 pixels
    for _ in range(10):
        manager.c.layout.grow()
    assert_dimensions(manager, 0, 150, 796, 446)

    # Min width is 25% of 600 = 150 pixels
    for _ in range(10):
        manager.c.layout.shrink()
    assert_dimensions(manager, 0, 450, 796, 146)


@monadtall_config
def test_tall_growsecondary_multiplesecondary(manager):
    manager.test_window("one")
    assert_dimensions(manager, 0, 0, 796, 596)

    manager.test_window("two")
    manager.test_window("three")
    manager.c.layout.previous()
    assert_focused(manager, "two")

    assert_dimensions(manager, 400, 0, 396, 296)
    # Grow 20 pixels
    manager.c.layout.grow()
    assert_dimensions(manager, 400, 0, 396, 316)
    manager.c.layout.shrink()
    assert_dimensions(manager, 400, 0, 396, 296)

    # Min height of other is 85 pixels, leaving 515
    for _ in range(20):
        manager.c.layout.grow()
    assert_dimensions(manager, 400, 0, 396, 511)

    # Min height of manager is 85 pixels
    for _ in range(40):
        manager.c.layout.shrink()
    assert_dimensions(manager, 400, 0, 396, 85)


@monadwide_config
def test_wide_growsecondary_multiplesecondary(manager):
    manager.test_window("one")
    assert_dimensions(manager, 0, 0, 796, 596)

    manager.test_window("two")
    manager.test_window("three")
    manager.c.layout.previous()
    assert_focused(manager, "two")

    assert_dimensions(manager, 0, 300, 396, 296)
    # Grow 20 pixels
    manager.c.layout.grow()
    assert_dimensions(manager, 0, 300, 416, 296)
    manager.c.layout.shrink()
    assert_dimensions(manager, 0, 300, 396, 296)

    # Min width of other is 85 pixels, leaving 715
    for _ in range(20):
        manager.c.layout.grow()
    assert_dimensions(manager, 0, 300, 710, 296)  # TODO why not 711 ?

    # Min width of manager is 85 pixels
    for _ in range(40):
        manager.c.layout.shrink()
    assert_dimensions(manager, 0, 300, 85, 296)


@monadtall_config
def test_tall_flip(manager):
    manager.test_window("one")
    manager.test_window("two")
    manager.test_window("three")

    # Check all the dimensions
    manager.c.layout.next()
    assert_focused(manager, "one")
    assert_dimensions(manager, 0, 0, 396, 596)

    manager.c.layout.next()
    assert_focused(manager, "two")
    assert_dimensions(manager, 400, 0, 396, 296)

    manager.c.layout.next()
    assert_focused(manager, "three")
    assert_dimensions(manager, 400, 300, 396, 296)

    # Now flip it and do it again
    manager.c.layout.flip()

    manager.c.layout.next()
    assert_focused(manager, "one")
    assert_dimensions(manager, 400, 0, 396, 596)

    manager.c.layout.next()
    assert_focused(manager, "two")
    assert_dimensions(manager, 0, 0, 396, 296)

    manager.c.layout.next()
    assert_focused(manager, "three")
    assert_dimensions(manager, 0, 300, 396, 296)


@monadwide_config
def test_wide_flip(manager):
    manager.test_window("one")
    manager.test_window("two")
    manager.test_window("three")

    # Check all the dimensions
    manager.c.layout.next()
    assert_focused(manager, "one")
    assert_dimensions(manager, 0, 0, 796, 296)

    manager.c.layout.next()
    assert_focused(manager, "two")
    assert_dimensions(manager, 0, 300, 396, 296)

    manager.c.layout.next()
    assert_focused(manager, "three")
    assert_dimensions(manager, 400, 300, 396, 296)

    # Now flip it and do it again
    manager.c.layout.flip()

    manager.c.layout.next()
    assert_focused(manager, "one")
    assert_dimensions(manager, 0, 300, 796, 296)

    manager.c.layout.next()
    assert_focused(manager, "two")
    assert_dimensions(manager, 0, 0, 396, 296)

    manager.c.layout.next()
    assert_focused(manager, "three")
    assert_dimensions(manager, 400, 0, 396, 296)


@monadtall_config
def test_tall_set_and_reset(manager):
    manager.test_window("one")
    assert_dimensions(manager, 0, 0, 796, 596)

    manager.test_window("two")
    assert_focused(manager, "two")
    assert_dimensions(manager, 400, 0, 396, 596)

    manager.c.layout.set_ratio(0.75)
    assert_focused(manager, "two")
    assert_dimensions(manager, 600, 0, 196, 596)

    manager.c.layout.set_ratio(0.25)
    assert_focused(manager, "two")
    assert_dimensions(manager, 200, 0, 596, 596)

    manager.c.layout.reset()
    assert_focused(manager, "two")
    assert_dimensions(manager, 400, 0, 396, 596)


@monadwide_config
def test_wide_set_and_reset(manager):
    manager.test_window("one")
    assert_dimensions(manager, 0, 0, 796, 596)

    manager.test_window("two")
    assert_focused(manager, "two")
    assert_dimensions(manager, 0, 300, 796, 296)

    manager.c.layout.set_ratio(0.75)
    assert_focused(manager, "two")
    assert_dimensions(manager, 0, 450, 796, 146)

    manager.c.layout.set_ratio(0.25)
    assert_focused(manager, "two")
    assert_dimensions(manager, 0, 150, 796, 446)

    manager.c.layout.reset()
    assert_focused(manager, "two")
    assert_dimensions(manager, 0, 300, 796, 296)


@monadtall_config
def test_tall_shuffle(manager):
    manager.test_window("one")
    manager.test_window("two")
    manager.test_window("three")
    manager.test_window("four")

    assert manager.c.layout.info()["main"] == "one"
    assert manager.c.layout.info()["secondary"] == ["two", "three", "four"]

    manager.c.layout.shuffle_up()
    assert manager.c.layout.info()["main"] == "one"
    assert manager.c.layout.info()["secondary"] == ["two", "four", "three"]

    manager.c.layout.shuffle_up()
    assert manager.c.layout.info()["main"] == "one"
    assert manager.c.layout.info()["secondary"] == ["four", "two", "three"]

    manager.c.layout.shuffle_up()
    assert manager.c.layout.info()["main"] == "four"
    assert manager.c.layout.info()["secondary"] == ["one", "two", "three"]


@monadwide_config
def test_wide_shuffle(manager):
    manager.test_window("one")
    manager.test_window("two")
    manager.test_window("three")
    manager.test_window("four")

    assert manager.c.layout.info()["main"] == "one"
    assert manager.c.layout.info()["secondary"] == ["two", "three", "four"]

    manager.c.layout.shuffle_up()
    assert manager.c.layout.info()["main"] == "one"
    assert manager.c.layout.info()["secondary"] == ["two", "four", "three"]

    manager.c.layout.shuffle_up()
    assert manager.c.layout.info()["main"] == "one"
    assert manager.c.layout.info()["secondary"] == ["four", "two", "three"]

    manager.c.layout.shuffle_up()
    assert manager.c.layout.info()["main"] == "four"
    assert manager.c.layout.info()["secondary"] == ["one", "two", "three"]


@monadtall_config
def test_tall_swap(manager):
    manager.test_window("one")
    manager.test_window("two")
    manager.test_window("three")
    manager.test_window("focused")

    assert manager.c.layout.info()["main"] == "one"
    assert manager.c.layout.info()["secondary"] == ["two", "three", "focused"]

    # Swap a secondary left, left aligned
    manager.c.layout.swap_left()
    assert manager.c.layout.info()["main"] == "focused"
    assert manager.c.layout.info()["secondary"] == ["two", "three", "one"]

    # Swap a main right, left aligned
    manager.c.layout.swap_right()
    assert manager.c.layout.info()["main"] == "two"
    assert manager.c.layout.info()["secondary"] == ["focused", "three", "one"]

    # flip over
    manager.c.layout.flip()
    manager.c.layout.shuffle_down()
    assert manager.c.layout.info()["main"] == "two"
    assert manager.c.layout.info()["secondary"] == ["three", "focused", "one"]

    # Swap secondary right, right aligned
    manager.c.layout.swap_right()
    assert manager.c.layout.info()["main"] == "focused"
    assert manager.c.layout.info()["secondary"] == ["three", "two", "one"]

    # Swap main left, right aligned
    manager.c.layout.swap_left()
    assert manager.c.layout.info()["main"] == "three"
    assert manager.c.layout.info()["secondary"] == ["focused", "two", "one"]

    # Do swap main
    manager.c.layout.swap_main()
    assert manager.c.layout.info()["main"] == "focused"
    assert manager.c.layout.info()["secondary"] == ["three", "two", "one"]

    # Since the focused window is already to the right this swap shouldn't
    # change the position of the windows
    # The swap function will try to get all windows to the right of the
    # focused window, which will result in a empty list that could cause
    # an error if not handled

    # Swap againts right edge
    manager.c.layout.swap_right()
    assert manager.c.layout.info()["main"] == "focused"
    assert manager.c.layout.info()["secondary"] == ["three", "two", "one"]

    # Same as above but for the swap_left function

    # Swap againts left edge
    manager.c.layout.swap_left()
    manager.c.layout.swap_left()
    assert manager.c.layout.info()["main"] == "three"
    assert manager.c.layout.info()["secondary"] == ["focused", "two", "one"]


@monadwide_config
def test_wide_swap(manager):
    manager.test_window("one")
    manager.test_window("two")
    manager.test_window("three")
    manager.test_window("focused")

    assert manager.c.layout.info()["main"] == "one"
    assert manager.c.layout.info()["secondary"] == ["two", "three", "focused"]

    # Swap a secondary up
    manager.c.layout.swap_right()  # equivalent to swap_down
    assert manager.c.layout.info()["main"] == "focused"
    assert manager.c.layout.info()["secondary"] == ["two", "three", "one"]

    # Swap a main down
    manager.c.layout.swap_left()  # equivalent to swap up
    assert manager.c.layout.info()["main"] == "two"
    assert manager.c.layout.info()["secondary"] == ["focused", "three", "one"]

    # flip over
    manager.c.layout.flip()
    manager.c.layout.shuffle_down()
    assert manager.c.layout.info()["main"] == "two"
    assert manager.c.layout.info()["secondary"] == ["three", "focused", "one"]

    # Swap secondary down
    manager.c.layout.swap_left()
    assert manager.c.layout.info()["main"] == "focused"
    assert manager.c.layout.info()["secondary"] == ["three", "two", "one"]

    # Swap main up
    manager.c.layout.swap_right()
    assert manager.c.layout.info()["main"] == "three"
    assert manager.c.layout.info()["secondary"] == ["focused", "two", "one"]

    # Do swap main
    manager.c.layout.swap_main()
    assert manager.c.layout.info()["main"] == "focused"
    assert manager.c.layout.info()["secondary"] == ["three", "two", "one"]

    # Since the focused window is already to the left this swap shouldn't
    # change the position of the windows
    # The swap function will try to get all windows to the left of the
    # focused window, which will result in a empty list that could cause
    # an error if not handled

    # Swap againts left edge
    manager.c.layout.swap_left()
    assert manager.c.layout.info()["main"] == "focused"
    assert manager.c.layout.info()["secondary"] == ["three", "two", "one"]

    # Same as above but for the swap_right function

    # Swap againts right edge
    manager.c.layout.swap_right()
    manager.c.layout.swap_right()
    assert manager.c.layout.info()["main"] == "three"
    assert manager.c.layout.info()["secondary"] == ["focused", "two", "one"]


@monadtall_config
def test_tall_window_focus_cycle(manager):
    # setup 3 tiled and two floating clients
    manager.test_window("one")
    manager.test_window("two")
    manager.test_window("float1")
    manager.c.window.toggle_floating()
    manager.test_window("float2")
    manager.c.window.toggle_floating()
    manager.test_window("three")

    # test preconditions
    assert manager.c.layout.info()["clients"] == ["one", "two", "three"]
    # last added window has focus
    assert_focused(manager, "three")

    # starting from the last tiled client, we first cycle through floating ones,
    # and afterwards through the tiled
    assert_focus_path(manager, "float1", "float2", "one", "two", "three")


@monadwide_config
def test_wide_window_focus_cycle(manager):
    # setup 3 tiled and two floating clients
    manager.test_window("one")
    manager.test_window("two")
    manager.test_window("float1")
    manager.c.window.toggle_floating()
    manager.test_window("float2")
    manager.c.window.toggle_floating()
    manager.test_window("three")

    # test preconditions
    assert manager.c.layout.info()["clients"] == ["one", "two", "three"]
    # last added window has focus
    assert_focused(manager, "three")

    # assert window focus cycle, according to order in layout
    assert_focus_path(manager, "float1", "float2", "one", "two", "three")


# MonadThreeCol
class MonadThreeColConfig(Config):
    auto_fullscreen = True
    groups = [libqtile.config.Group("a")]
    layouts = [layout.MonadThreeCol()]
    floating_layout = libqtile.resources.default_config.floating_layout
    keys = []
    mouse = []
    screens = []
    follow_mouse_focus = False


monadthreecol_config = pytest.mark.parametrize("manager", [MonadThreeColConfig], indirect=True)


@monadthreecol_config
def test_three_col_add_clients(manager):
    manager.test_window("one")
    assert manager.c.layout.info()["main"] == "one"
    assert manager.c.layout.info()["secondary"] == dict(left=[], right=[])

    manager.test_window("two")
    assert manager.c.layout.info()["main"] == "two"
    assert manager.c.layout.info()["secondary"] == dict(left=["one"], right=[])
    assert_focused(manager, "two")

    manager.test_window("three")
    assert manager.c.layout.info()["main"] == "three"
    assert manager.c.layout.info()["secondary"] == dict(left=["two"], right=["one"])
    assert_focused(manager, "three")

    manager.test_window("four")
    assert manager.c.layout.info()["main"] == "four"
    assert manager.c.layout.info()["secondary"] == dict(left=["three", "two"], right=["one"])
    assert_focused(manager, "four")

    manager.test_window("five")
    assert manager.c.layout.info()["main"] == "five"
    assert manager.c.layout.info()["secondary"] == dict(
        left=["four", "three"], right=["two", "one"]
    )
    assert_focused(manager, "five")

    manager.c.layout.next()
    assert_focused(manager, "four")
    manager.c.layout.next()
    assert_focused(manager, "three")
    manager.c.layout.next()
    assert_focused(manager, "two")
    manager.c.layout.next()
    assert_focused(manager, "one")


@monadthreecol_config
def test_three_col_shuffle(manager):
    manager.test_window("one")
    manager.test_window("two")
    manager.test_window("three")
    manager.test_window("four")
    manager.test_window("five")

    manager.c.layout.shuffle_right()
    assert manager.c.layout.info()["main"] == "two"
    assert manager.c.layout.info()["secondary"] == dict(
        left=["four", "three"], right=["five", "one"]
    )
    assert_focused(manager, "five")

    manager.c.layout.shuffle_down()
    assert manager.c.layout.info()["main"] == "two"
    assert manager.c.layout.info()["secondary"] == dict(
        left=["four", "three"], right=["one", "five"]
    )
    assert_focused(manager, "five")

    manager.c.layout.shuffle_left()
    assert manager.c.layout.info()["main"] == "five"
    assert manager.c.layout.info()["secondary"] == dict(
        left=["four", "three"], right=["one", "two"]
    )
    assert_focused(manager, "five")

    manager.c.layout.shuffle_left()
    assert manager.c.layout.info()["main"] == "four"
    assert manager.c.layout.info()["secondary"] == dict(
        left=["five", "three"], right=["one", "two"]
    )
    assert_focused(manager, "five")

    manager.c.layout.shuffle_down()
    assert manager.c.layout.info()["main"] == "four"
    assert manager.c.layout.info()["secondary"] == dict(
        left=["three", "five"], right=["one", "two"]
    )
    assert_focused(manager, "five")

    manager.c.layout.shuffle_up()
    assert manager.c.layout.info()["main"] == "four"
    assert manager.c.layout.info()["secondary"] == dict(
        left=["five", "three"], right=["one", "two"]
    )
    assert_focused(manager, "five")

    manager.c.layout.shuffle_right()
    assert manager.c.layout.info()["main"] == "five"
    assert manager.c.layout.info()["secondary"] == dict(
        left=["four", "three"], right=["one", "two"]
    )
    assert_focused(manager, "five")


@monadthreecol_config
def test_three_col_swap_main(manager):
    manager.test_window("one")
    manager.test_window("two")
    manager.test_window("three")
    manager.test_window("four")
    manager.test_window("five")

    manager.c.layout.next()
    manager.c.layout.swap_main()
    assert manager.c.layout.info()["main"] == "four"
    assert manager.c.layout.info()["secondary"] == dict(
        left=["five", "three"], right=["two", "one"]
    )
    assert_focused(manager, "four")

    manager.c.layout.next()
    manager.c.layout.swap_main()
    assert manager.c.layout.info()["main"] == "five"
    assert manager.c.layout.info()["secondary"] == dict(
        left=["four", "three"], right=["two", "one"]
    )
    assert_focused(manager, "five")



# Copyright (c) 2021 elParaguayo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Widget specific tests

import sys
from importlib import reload
from types import ModuleType

from libqtile.widget import gmail_checker
from test.widgets.conftest import FakeBar


class FakeIMAP(ModuleType):
    class IMAP4_SSL:  # noqa: N801
        def __init__(self, *args, **kwargs):
            pass

        def login(self, username, password):
            self.username = username
            self.password = password

        def status(self, path, *args, **kwargs):
            if not (self.username and self.password):
                return False, None

            return ("OK", ['("{}" (MESSAGES 10 UNSEEN 2)'.format(path).encode()])


def test_gmail_checker_valid_response(fake_qtile, monkeypatch, fake_window):
    monkeypatch.setitem(sys.modules, "imaplib", FakeIMAP("imaplib"))
    reload(gmail_checker)

    gmc = gmail_checker.GmailChecker(username="qtile", password="test")
    fakebar = FakeBar([gmc], window=fake_window)
    gmc._configure(fake_qtile, fakebar)
    text = gmc.poll()
    assert text == "inbox[10],unseen[2]"


def test_gmail_checker_invalid_response(fake_qtile, monkeypatch, fake_window):
    monkeypatch.setitem(sys.modules, "imaplib", FakeIMAP("imaplib"))
    reload(gmail_checker)

    gmc = gmail_checker.GmailChecker()
    fakebar = FakeBar([gmc], window=fake_window)
    gmc._configure(fake_qtile, fakebar)
    text = gmc.poll()
    assert text == "UNKNOWN ERROR"


# This test is only required because the widget is written
# inefficiently. display_fmt should use keys instead of indices.
def test_gmail_checker_only_unseen(fake_qtile, monkeypatch, fake_window):
    monkeypatch.setitem(sys.modules, "imaplib", FakeIMAP("imaplib"))
    reload(gmail_checker)

    gmc = gmail_checker.GmailChecker(
        display_fmt="unseen[{0}]", status_only_unseen=True, username="qtile", password="test"
    )
    fakebar = FakeBar([gmc], window=fake_window)
    gmc._configure(fake_qtile, fakebar)
    text = gmc.poll()
    assert text == "unseen[2]"



# Copyright (c) 2020 Tycho Andersen
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import os
import platform
import shutil
import subprocess
import tempfile
import textwrap

import pytest


def is_cpython():
    # https://mypy.readthedocs.io/en/stable/faq.html#does-it-run-on-pypy
    return platform.python_implementation() == "CPython"


def have_mypy():
    return shutil.which("mypy") is not None


pytestmark = pytest.mark.skipif(not is_cpython() or not have_mypy(), reason="needs mypy")


def run_qtile_check(config):
    # make sure we have mypy so we can type check the config
    if shutil.which("mypy") is None:
        raise Exception("missing mypy in test environment")
    cmd = os.path.join(os.path.dirname(__file__), "..", "bin", "qtile")
    argv = [cmd, "check", "-c", config]
    try:
        newenv = os.environ.copy()
        old_pp = newenv.get("PYTHONPATH", "")
        newenv["PYTHONPATH"] = os.path.join(os.path.dirname(__file__), "..") + ":" + old_pp
        subprocess.check_call(argv, env=newenv)
    except subprocess.CalledProcessError:
        return False
    else:
        return True


def test_check_default_config():
    # the default config should always be ok
    default_config = os.path.join(
        os.path.dirname(__file__), "..", "libqtile", "resources", "default_config.py"
    )
    assert run_qtile_check(default_config)


def check_literal_config(config):
    with tempfile.TemporaryDirectory() as tempdir:
        with tempfile.NamedTemporaryFile(dir=tempdir, suffix=".py") as temp:
            temp.write(textwrap.dedent(config).encode("utf-8"))
            temp.flush()
            return run_qtile_check(temp.name)


def test_check_bad_syntax():
    assert not check_literal_config(
        """
        meshuggah rocks!
    """
    )


def test_check_bad_key_arg():
    assert not check_literal_config(
        """
            Key([1], 1)
    """
    )


def test_check_good_key_arg():
    assert check_literal_config(
        """
        from libqtile.config import Key
        from libqtile.command import lazy

        keys = [Key(["mod4"], "x", lazy.kill())]
        cursor_warp = True
    """
    )


def test_check_bad_config_type():
    assert not check_literal_config(
        """
        keys = "i am not a List[Key]"
        cursor_warp = "i am not a bool"
    """
    )


def test_extra_vars_are_ok():
    assert check_literal_config(
        """
        this_is_an_extra_config_var = "yes it is"
    """
    )


def test_extra_files_are_ok():
    with tempfile.TemporaryDirectory() as tempdir:
        config_file = os.path.join(tempdir, "config.py")
        with open(config_file, "w") as config:
            config.write("from bar import foo\n")
        with open(os.path.join(tempdir, "bar.py"), "w") as config:
            config.write("foo = 42")
        assert run_qtile_check(config_file)



# Copyright (c) 2021 elParaguayo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import libqtile.bar
import libqtile.config
import libqtile.confreader
import libqtile.layout
from libqtile import widget
from libqtile.lazy import lazy


def test_lazy_callback(manager_nospawn, minimal_conf_noscreen):
    """Test widgets accept lazy calls"""
    textbox = widget.TextBox(
        text="Testing",
        mouse_callbacks={
            "Button1": lazy.widget["textbox"].update("LazyCall"),
        },
    )

    config = minimal_conf_noscreen
    config.screens = [libqtile.config.Screen(top=libqtile.bar.Bar([textbox], 10))]

    manager_nospawn.start(config)

    topbar = manager_nospawn.c.bar["top"]
    assert topbar.widget["textbox"].info()["text"] == "Testing"

    topbar.fake_button_press(0, "top", 0, 0, button=1)
    assert topbar.widget["textbox"].info()["text"] == "LazyCall"
